import textwrap
import unittest

from brew.constants import GRAIN_TYPE_CEREAL
from brew.constants import GRAIN_TYPE_DME
from brew.constants import GRAIN_TYPE_LME
from brew.constants import GRAIN_TYPE_SPECIALTY
from brew.constants import IMPERIAL_UNITS
from brew.constants import SI_UNITS
from brew.grains import Grain
from brew.grains import GrainAddition
from fixtures import pale
from fixtures import pale_add


class TestGrains(unittest.TestCase):

    def setUp(self):
        self.grain = pale

    def test_str(self):
        out = str(self.grain)
        self.assertEquals(out, 'Pale 2-row')

    def test_repr(self):
        out = repr(self.grain)
        self.assertEquals(out, "Grain('pale 2-row', short_name='2-row', color=2.0, hwe=308.78)")  # nopep8

    def test_format(self):
        out = self.grain.format()
        msg = textwrap.dedent("""\
            pale 2-row Grain
            -----------------------------------
            Color:             2.0 degL
            PPG:               37.00
            Hot Water Extract: 308.78""")
        self.assertEquals(out, msg)

    def test_get_working_yield(self):
        wy = self.grain.get_working_yield(0.70)
        self.assertEquals(round(wy, 2), 0.56)

    def test_grain_hwe(self):
        pale = Grain('pale 2-row',
                     short_name='2-row',
                     color=2.0,
                     hwe=308.0)
        self.assertEquals(pale.hwe, 308.0)
        self.assertEquals(round(pale.ppg, 2), 36.91)

    def test_grain_ppg_hwe_raises(self):
        with self.assertRaises(Exception):
            Grain('pale 2-row',
                  short_name='2-row',
                  color=2.0,
                  ppg=37.0,
                  hwe=308.0)

    def test_to_dict(self):
        out = self.grain.to_dict()
        expected = {'name': 'pale 2-row',
                    'short_name': '2-row',
                    'color': 2.0,
                    'ppg': 37.0,
                    'hwe': 308.78,
                    }
        self.assertEquals(out, expected)

    def test_to_json(self):
        out = self.grain.to_json()
        expected = '{"color": 2.0, "hwe": 308.78, "name": "pale 2-row", "ppg": 37.0, "short_name": "2-row"}'  # nopep8
        self.assertEquals(out, expected)


class TestGrainAdditions(unittest.TestCase):

    def setUp(self):
        self.grain_add = pale_add

    def test_str(self):
        out = str(self.grain_add)
        self.assertEquals(out, 'Pale 2-row, weight 13.96 lbs')

    def test_repr(self):
        out = repr(self.grain_add)
        self.assertEquals(out, "GrainAddition(Grain('pale 2-row', short_name='2-row', color=2.0, hwe=308.78), weight=13.96, grain_type='cereal')")  # nopep8

    def test_get_weight_cereal(self):
        grain_add = GrainAddition(pale, weight=13.96,
                                  grain_type=GRAIN_TYPE_CEREAL)
        self.assertEquals(grain_add.grain_type, GRAIN_TYPE_CEREAL)
        out = grain_add.get_cereal_weight()
        self.assertEqual(round(out, 2), 13.96)
        out = grain_add.get_lme_weight()
        self.assertEqual(round(out, 2), 10.47)
        out = grain_add.get_dry_weight()
        self.assertEqual(round(out, 2), 8.38)

    def test_get_weight_lme(self):
        grain_add = GrainAddition(pale, weight=10.47,
                                  grain_type=GRAIN_TYPE_LME)
        self.assertEquals(grain_add.grain_type, GRAIN_TYPE_LME)
        out = grain_add.get_cereal_weight()
        self.assertEqual(round(out, 2), 13.96)
        out = grain_add.get_lme_weight()
        self.assertEqual(round(out, 2), 10.47)
        out = grain_add.get_dry_weight()
        self.assertEqual(round(out, 2), 8.38)

    def test_get_weight_dry(self):
        grain_add = GrainAddition(pale, weight=8.38,
                                  grain_type=GRAIN_TYPE_DME)
        self.assertEquals(grain_add.grain_type, GRAIN_TYPE_DME)
        out = grain_add.get_cereal_weight()
        self.assertEqual(round(out, 2), 13.97)
        out = grain_add.get_lme_weight()
        self.assertEqual(round(out, 2), 10.48)
        out = grain_add.get_dry_weight()
        self.assertEqual(round(out, 2), 8.38)

    def test_get_weight_specialty(self):
        grain_add = GrainAddition(pale, weight=13.96,
                                  grain_type=GRAIN_TYPE_SPECIALTY)
        self.assertEquals(grain_add.grain_type, GRAIN_TYPE_SPECIALTY)
        out = grain_add.get_cereal_weight()
        self.assertEqual(round(out, 2), 13.96)
        out = grain_add.get_lme_weight()
        self.assertEqual(round(out, 2), 12.42)
        out = grain_add.get_dry_weight()
        self.assertEqual(round(out, 2), 9.94)

    def test_get_weight_map(self):
        out = self.grain_add.get_weight_map()
        expected = {
            'grain_weight': 13.96,
            'lme_weight': 10.47,
            'dry_weight': 8.38,
        }
        self.assertEquals(out, expected)

    def test_to_dict(self):
        out = self.grain_add.to_dict()
        expected = {'name': 'pale 2-row',
                    'data': {
                        'short_name': '2-row',
                        'color': 2.0,
                        'ppg': 37.0,
                        'hwe': 308.78,
                    },
                    'grain_type': 'cereal',
                    'weight': 13.96,
                    'units': 'imperial',
                    }
        self.assertEquals(out, expected)

    def test_to_json(self):
        out = self.grain_add.to_json()
        expected = '{"data": {"color": 2.0, "hwe": 308.78, "ppg": 37.0, "short_name": "2-row"}, "grain_type": "cereal", "name": "pale 2-row", "units": "imperial", "weight": 13.96}'  # nopep8
        self.assertEquals(out, expected)

    def test_validate(self):
        data = self.grain_add.to_dict()
        GrainAddition.validate(data)

    def test_format(self):
        out = self.grain_add.format()
        msg = textwrap.dedent("""\
            pale 2-row Addition
            -----------------------------------
            Grain Type:        cereal
            Weight:            13.96 lbs""")
        self.assertEquals(out, msg)

    def test_grain_change_units_imperial_to_si(self):
        self.assertEquals(self.grain_add.units, IMPERIAL_UNITS)
        self.assertEquals(round(self.grain_add.weight, 2), 13.96)
        grain_add = self.grain_add.change_units()
        self.assertEquals(grain_add.units, SI_UNITS)
        self.assertEquals(round(grain_add.weight, 2), 6.33)

    def test_grain_change_units_si_to_imperial(self):
        grain_add = self.grain_add.change_units()
        self.assertEquals(grain_add.units, SI_UNITS)
        self.assertEquals(round(grain_add.weight, 2), 6.33)
        grain_add = grain_add.change_units()
        self.assertEquals(grain_add.units, IMPERIAL_UNITS)
        self.assertEquals(round(grain_add.weight, 2), 13.96)



# Unit Types
IMPERIAL_UNITS = 'imperial'
SI_UNITS = 'metric'

IMPERIAL_TYPES = {
    'volume': 'gallon',
    'weight_large': 'lbs',
    'weight_small': 'oz',
    'temperature': 'degF',
}

SI_TYPES = {
    'volume': 'liter',
    'weight_large': 'kg',
    'weight_small': 'mg',
    'temperature': 'degC',
}

# Grains

GRAIN_TYPE_CEREAL = 'cereal'
GRAIN_TYPE_SPECIALTY = 'specialty'
GRAIN_TYPE_DME = 'dme'
GRAIN_TYPE_LME = 'lme'
GRAIN_TYPE_LIST = [
    GRAIN_TYPE_CEREAL,
    GRAIN_TYPE_SPECIALTY,
    GRAIN_TYPE_DME,
    GRAIN_TYPE_LME,
]

# Hops

HOP_TYPE_PELLET = 'pellet'
HOP_TYPE_WHOLE = 'whole'
HOP_TYPE_PLUG = 'plug'
HOP_TYPE_LIST = [
    HOP_TYPE_PELLET,
    HOP_TYPE_WHOLE,
    HOP_TYPE_PLUG,
]

# Hop utilization scale factors
HOP_UTILIZATION_SCALE_PELLET = 1.1  # 110%

# Conversions
# Try to define SI to US units instead of US to SI.
MG_PER_OZ = 28349.5
OZ_PER_MG = 1.0 / MG_PER_OZ

G_PER_OZ = 28.3495
OZ_PER_G = 1.0 / G_PER_OZ

KG_PER_POUND = 0.453592
POUND_PER_KG = 1.0 / KG_PER_POUND

LITER_PER_GAL = 3.78541
GAL_PER_LITER = 1.0 / LITER_PER_GAL

# 1 oz/gallon = 7489.15 mg/l
# Use this when calculating IBUS
HOPS_CONSTANT_IMPERIAL = MG_PER_OZ * GAL_PER_LITER
HOPS_CONSTANT_SI = 1.0

# Elevation
SEA_LEVEL = 0     # feet or meters
MILE_HIGH = 5280  # feet
KM_HIGH = 1000    # meters

# Grind Constants
# fine/coarse difference percentage for different grain in decimal format
FC_DIFF_TWO_ROW = 0.017
FC_DIFF_SIX_ROW = 0.015

# Moisture in finished malt as percentage in decimal form (ie 4% is 0.04)
MOISTURE_FINISHED_MALT = 0.04
# I've seen this correction but haven't found a source, set to zero instead
# MOISTURE_CORRECTION = 0.002
MOISTURE_CORRECTION = 0.0

# Sucrose is considered 100% extractable in water, so the maximum PPG and
# Plato are listed here
SUCROSE_PLATO = 11.486
SUCROSE_PPG = 46.214

# Hot Water Extract is a measure of how many liters of wort are required at a
# Specific Gravity of 1.001
LITERS_OF_WORT_AT_SG = 386.0
PPG_TO_HWE_CONVERSION = LITER_PER_GAL * POUND_PER_KG

# Most hydrometers are calibrated to 59 degF
HYDROMETER_ADJUSTMENT_TEMP = 59.0

# Weight of Water
WATER_WEIGHT_IMPERIAL = 8.32  # lbs / gallon
WATER_WEIGHT_SI = 1.0  # kg / liter

# Alcohol
ALCOHOL_SPECIFIC_GRAVITY = 0.7936



from brew.grains import Grain
from brew.grains import GrainAddition
from brew.hops import Hop
from brew.hops import HopAddition
from brew.recipes import Recipe
from brew.yeasts import Yeast

# Define Grains
pale = Grain('pale 2-row',
             short_name='2-row',
             color=2.0,
             ppg=37.0)
crystal = Grain('crystal C20',
                short_name='C20',
                color=20.0,
                ppg=35.0)
grain_list = [pale, crystal]

pale_add = GrainAddition(pale,
                         weight=13.96)

crystal_add = GrainAddition(crystal,
                            weight=0.78)
grain_additions = [pale_add, crystal_add]

# Define Hops
centennial = Hop(name='centennial',
                 percent_alpha_acids=0.14)
cascade = Hop(name='cascade',
              percent_alpha_acids=0.07)
hop_list = [centennial, cascade]

centennial_add = HopAddition(centennial,
                             boil_time=60.0,
                             weight=0.57)
cascade_add = HopAddition(cascade,
                          boil_time=5.0,
                          weight=0.76)
hop_additions = [centennial_add, cascade_add]

# Define Yeast
yeast = Yeast("Wyeast 1056")

# Define Recipes
recipe = Recipe(name='pale ale',
                grain_additions=grain_additions,
                hop_additions=hop_additions,
                yeast=yeast,
                percent_brew_house_yield=0.70,  # %
                start_volume=7.0,  # G
                final_volume=5.0,  # G
                )



import glob
import json
import os

from brew.grains import Grain
from brew.grains import GrainAddition
from brew.hops import Hop
from brew.hops import HopAddition
from brew.recipes import Recipe
from brew.utilities.sugar import sg_to_gu
from brew.yeasts import Yeast


class DataLoader(object):
    """
    Base class for loading data from data files inside the data_dir.
    """
    DATA = {}
    EXT = ''

    def __init__(self, data_dir):
        self.data_dir = data_dir

    @classmethod
    def format_name(cls, name):
        """
        Reformat a given name to match the filename of a data file.
        """
        return name.lower().replace(' ', '_').replace('-', '_')

    @classmethod
    def read_data(cls, filename):
        raise NotImplementedError

    def get_item(self, dir_suffix, item_name):
        item_dir = os.path.join(self.data_dir, dir_suffix)

        # Cache the directory
        if dir_suffix not in self.DATA:
            self.DATA[dir_suffix] = {}
            for item in glob.glob('{}*.{}'.format(item_dir, self.EXT)):
                ext_len = len(self.EXT) + 1
                filename = os.path.basename(item)[:-ext_len]
                self.DATA[dir_suffix][filename] = {}

        name = self.format_name(item_name)
        if name not in self.DATA[dir_suffix]:
            raise Exception('Item from {} dir not found: {}'.format(dir_suffix,
                                                                    name))

        # Cache file data
        if not self.DATA[dir_suffix][name]:
            item_filename = os.path.join(item_dir, '{}.{}'.format(name, self.EXT))  # nopep8
            data = self.read_data(item_filename)
            self.DATA[dir_suffix][name] = data
            return data
        else:
            return self.DATA[dir_suffix][name]


class JSONDataLoader(DataLoader):
    """
    Load data from JSON files inside the data_dir.
    """
    EXT = 'json'

    @classmethod
    def read_data(cls, filename):
        data = None
        with open(filename, 'r') as data_file:
            data = json.loads(data_file.read())
        return data


def parse_cereals(cereal, loader):
    """
    Parse grains data from a recipe

    Grain must have the following top level attributes:
    - name   (str)
    - weight (float)
    - data   (dict) (optional)

    Additionally grains may contain override data in the 'data'
    attribute with the following keys:
    - color (float)
    - ppg   (int)
    """
    GrainAddition.validate(cereal)

    cereal_data = {}
    try:
        cereal_data = loader.get_item('cereals/', cereal['name'])
    except Exception:
        pass

    name = cereal_data.get('name', cereal['name'])
    color = None
    ppg = None

    if 'data' in cereal:
        color = cereal['data'].get('color', None)
        ppg = cereal['data'].get('ppg', None)

    if not color:
        color = float(cereal_data['color'][:-4])

    if not ppg:
        ppg = sg_to_gu(float(cereal_data['potential'][:-3]))

    grain_obj = Grain(name, color=color, ppg=ppg)

    grain_add_kwargs = {
        'weight': float(cereal['weight']),
    }
    if 'grain_type' in cereal:
        grain_add_kwargs['grain_type'] = cereal['grain_type']
    if 'units' in cereal:
        grain_add_kwargs['units'] = cereal['units']
    return GrainAddition(grain_obj, **grain_add_kwargs)


def parse_hops(hop, loader):
    """
    Parse hops data from a recipe

    Hops must have the following top level attributes:
    - name      (str)
    - weight    (float)
    - boil_time (float)
    - data      (dict) (optional)

    Additionally hops may contain override data in the 'data' attribute
    with the following keys:
    - percent_alpha_acids (float)
    """
    HopAddition.validate(hop)

    hop_data = {}
    try:
        hop_data = loader.get_item('hops/', hop['name'])
    except Exception:
        pass

    name = hop_data.get('name', hop['name'])
    alpha_acids = None

    if 'data' in hop:
        alpha_acids = hop['data'].get('percent_alpha_acids', None)

    if not alpha_acids:
        alpha_acids = float(hop_data['alpha_acid_composition'].split('%')[0]) / 100.  # nopep8

    hop_obj = Hop(name, percent_alpha_acids=alpha_acids)
    hop_add_kwargs = {
        'weight': float(hop['weight']),
        'boil_time': hop['boil_time'],
    }
    if 'hop_type' in hop:
        hop_add_kwargs['hop_type'] = hop['hop_type']
    if 'units' in hop:
        hop_add_kwargs['units'] = hop['units']
    return HopAddition(hop_obj, **hop_add_kwargs)


def parse_yeast(yeast, loader):
    """
    Parse yeast data from a recipe

    Yeast must have the following top level attributes:
    - name (str)
    - data (dict) (optional)

    Additionally yeast may contain override data in the 'data' attribute
    with the following keys:
    - percent_attenuation (float)
    """
    Yeast.validate(yeast)

    yeast_data = {}
    try:
        yeast_data = loader.get_item('yeast/', yeast['name'])  # nopep8
    except Exception:
        return Yeast(yeast['name'])

    name = yeast_data.get('name', yeast['name'])
    attenuation = None

    if 'data' in yeast:
        attenuation = yeast['data'].get('percent_attenuation', None)

    if not attenuation:
        attenuation = yeast_data['attenuation'][0]

    return Yeast(name, percent_attenuation=attenuation)


def parse_recipe(recipe, loader):
    """
    Parse a recipe from a python Dict

    recipe: a python dict describing the recipe
    loader: a data loader class that loads data from data files

    A recipe must have the following top level attributes:
    - name         (str)
    - start_volume (float)
    - final_volume (float)
    - grains       (list(dict))
    - hops         (list(dict))
    - yeast        (dict)

    Additionally the recipe may contain override data in the 'data'
    attribute with the following keys:
    - percent_brew_house_yield (float)
    - units                    (str)

    All other fields will be ignored and may be used for other metadata.

    The dict objects in the grains, hops, and yeast values are required to have
    the key 'name' and the remaining attributes will be looked up in the data
    directory if they are not provided.
    """
    Recipe.validate(recipe)

    grain_additions = []
    for grain in recipe['grains']:
        grain_additions.append(parse_cereals(grain, loader))

    hop_additions = []
    for hop in recipe['hops']:
        hop_additions.append(parse_hops(hop, loader))

    yeast = parse_yeast(recipe['yeast'], loader)

    recipe_kwargs = {
        'grain_additions': grain_additions,
        'hop_additions': hop_additions,
        'yeast': yeast,
        'start_volume': recipe['start_volume'],
        'final_volume': recipe['final_volume'],
    }
    if 'data' in recipe:
        if 'percent_brew_house_yield' in recipe['data']:
            recipe_kwargs['percent_brew_house_yield'] = \
                recipe['data']['percent_brew_house_yield']
        if 'units' in recipe['data']:
            recipe_kwargs['units'] = recipe['data']['units']

    beer = Recipe(recipe['name'],
                  **recipe_kwargs)
    return beer



#! /usr/bin/env python

"""
Brewing Classic Styles: 80 Winning Recipes Anyone Can Brew
by Jamil Zainasheff and John J. Palmer

Munich Madness

Used by permission of Brewers Publications (2007). All rights reserved.

You can purchase the book here:

- http://www.brewerspublications.com/books/brewing-classic-styles-80-winning-recipes-anyone-can-brew/

Original Stats:

OG:      1.055 (13.6P)
FG:      1.015 ( 3.7P)
ADF:     73%
IBU:     27
Color:   11 SRM (21 EBC)
Alcohol: 5.4% ABV (4.2% ABW)
Boil:    60 min
Pre-Boil Volume:  7 Gal (26.5L)
Pre-Boil Gravity: 1.047 (11.7P)
"""  # nopep8

import os

from brew.parsers import JSONDataLoader
from brew.parsers import parse_recipe


def main():

    recipe = {
        'name': "Munich Madness (All Grain)",
        'start_volume': 7.0,
        'final_volume': 6.0,
        'grains': [
            {'name': 'Pilsner 2 row Ger',
             'data': {
                 'color': 2.3,
             },
             'weight': 5.0},
            {'name': 'Munich Malt 10L',
             'data': {
                 'color': 9.0,
             },
             'weight': 4.0},
            {'name': 'Vienna Malt',
             'weight': 3.0},
            {'name': 'Caramunich Malt',
             'data': {
                 'color': 60.0,
             },
             'weight': 1.0,
             'grain_type': 'specialty'},
        ],
        'hops': [
            {'name': 'Hallertau US',
             'data': {
                 'percent_alpha_acids': 0.04,
             },
             'weight': 1.5,
             'boil_time': 60.0},
            {'name': 'Hallertau US',
             'data': {
                 'percent_alpha_acids': 0.04,
             },
             'weight': 0.5,
             'boil_time': 20.0},
        ],
        'yeast': {
            'name': 'Wyeast 2206',
            'data': {
                'percent_attenuation': 0.73,
            },
        },
        'data': {
            'percent_brew_house_yield': 0.70,
            'units': 'imperial',
        },
    }

    data_dir = os.path.abspath(os.path.join(os.getcwd(), 'data/'))
    loader = JSONDataLoader(data_dir)
    beer = parse_recipe(recipe, loader)
    print(beer.format())


if __name__ == "__main__":
    main()



import math

from ..constants import HOP_TYPE_PELLET
from ..constants import HOP_UTILIZATION_SCALE_PELLET
from ..constants import HOPS_CONSTANT_IMPERIAL
from ..constants import HOPS_CONSTANT_SI
from ..constants import IMPERIAL_TYPES
from ..constants import IMPERIAL_UNITS
from ..constants import SI_TYPES
from ..constants import SI_UNITS
from ..validators import validate_units


__all__ = [
    'HopsUtilization',
    'HopsUtilizationJackieRager',
    'HopsUtilizationGlennTinseth',
]


class HopsUtilization(object):
    """
    http://www.boondocks-brewing.com/hops
    """

    def __init__(self, hop_addition,
                 units=IMPERIAL_UNITS):
        self.hop_addition = hop_addition

        # Manage units
        self.set_units(units)

    def set_units(self, units):
        self.units = validate_units(units)
        if self.units == IMPERIAL_UNITS:
            self.types = IMPERIAL_TYPES
        elif self.units == SI_UNITS:
            self.types = SI_TYPES

    def change_units(self):
        """
        Change units from one type to the other return new instance
        """
        if self.units == IMPERIAL_UNITS:
            units = SI_UNITS
        elif self.units == SI_UNITS:
            units = IMPERIAL_UNITS
        return HopsUtilization(self.hop_addition,
                               units=units)

    def get_ibus(self, sg, final_volume):
        hops_constant = HOPS_CONSTANT_IMPERIAL
        if self.units == SI_UNITS:
            hops_constant = HOPS_CONSTANT_SI
        utilization = self.get_percent_utilization(
            sg, self.hop_addition.boil_time)
        # Utilization is 10% higher for pellet vs whole/plug
        if self.hop_addition.hop_type == HOP_TYPE_PELLET:
            utilization *= HOP_UTILIZATION_SCALE_PELLET
        num = (self.hop_addition.weight * utilization *
               self.hop_addition.hop.percent_alpha_acids *
               hops_constant)
        return num / final_volume

    @classmethod
    def get_percent_utilization(cls, sg, boil_time):
        raise NotImplementedError

    @classmethod
    def get_utilization_table(cls, gravity_list, boil_time_list, sig=3):
        table = []
        for boil_time in boil_time_list:
            line = []
            for sg in gravity_list:
                aau = cls.get_percent_utilization(sg / 1000.0, boil_time)
                line.append(round(aau, sig))
            table.append(line)
        return table

    @classmethod
    def format_utilization_table(cls):
        """
        Percent Alpha Acid Utilization - Boil Time vs Wort Original Gravity

        Source: http://www.realbeer.com/hops/research.html
        """
        gravity_list = list(range(1030, 1140, 10))
        boil_time_list = list(range(0, 60, 3)) + list(range(60, 130, 10))
        table = cls.get_utilization_table(gravity_list, boil_time_list)

        title = 'Percent Alpha Acid Utilization - ' \
                'Boil Time vs Wort Original Gravity'
        table_size = 92

        out = []
        out.append(title.center(table_size))
        out.append(str('=' * len(title)).center(table_size))
        out.append('\n')
        out.append(' '.join([' ' * 4] + ['{0:7.3f}'.format(l / 1000.0)
                   for l in gravity_list]))
        out.append('-' * table_size)
        for index, line in enumerate(table):
            boil_time = boil_time_list[index]
            out.append('{0} {1}'.format(
                str(boil_time).rjust(4),
                ' '.join(['{0:7.3f}'.format(aau) for aau in line])))
        return '\n'.join([o.rstrip() for o in out if o != '\n'])


class HopsUtilizationJackieRager(HopsUtilization):
    """
    Jackie Rager

    Best for extract and partial mash brewing.

    Source: http://www.rooftopbrew.net/ibu.php
    """

    def __str__(self):
        return "Jackie Rager"

    @classmethod
    def get_c_gravity(cls, sg):
        """
        Cgravity is a constant to adjust the boil size when dealing with
        specific gravity greater than 1.050 in the calculation of IBUs.
        """
        cgravity = 1
        if sg > 1.050:
            cgravity += (sg - 1.050) / 0.2
        return cgravity

    @classmethod
    def get_percent_utilization(cls, sg, boil_time):
        num = (18.11 + 13.86 * math.tanh((boil_time - 31.32) / 18.27)) / 100.0
        return num / cls.get_c_gravity(sg)


class HopsUtilizationGlennTinseth(HopsUtilization):
    """
    Glenn Tinseth

    Best for all grain brewing.

    Source: http://www.realbeer.com/hops/research.html
    Source: http://www.rooftopbrew.net/ibu.php
    """

    def __str__(self):
        return "Glenn Tinseth"

    @classmethod
    def get_bigness_factor(cls, sg):
        return 1.65 * 0.000125 ** (sg - 1)

    @classmethod
    def get_boil_time_factor(cls, boil_time):
        return (1 - math.exp(-0.04 * boil_time)) / 4.15

    @classmethod
    def get_percent_utilization(cls, sg, boil_time):
        """
        The Bigness factor accounts for reduced utilization due to higher wort
        gravities. Use an average gravity value for the entire boil to account
        for changes in the wort volume.

        Bigness factor = 1.65 * 0.000125^(wort gravity - 1)

        The Boil Time factor accounts for the change in utilization due to
        boil time:

        Boil Time factor = (1 - e^(-0.04 * time in mins)) / 4.15

        Source: http://www.realbeer.com/hops/research.html
        """
        bigness_factor = cls.get_bigness_factor(sg)
        boil_time_factor = cls.get_boil_time_factor(boil_time)
        return bigness_factor * boil_time_factor



import unittest

from brew.constants import IMPERIAL_UNITS
from brew.constants import SI_UNITS
from brew.utilities.yeast import KaiserYeastModel
from brew.utilities.yeast import WhiteYeastModel
from brew.utilities.yeast import YeastModel
from brew.utilities.yeast import pitch_rate_conversion


class TestYeastUtilities(unittest.TestCase):

    def test_pitch_rate_conversion_si(self):
        out = pitch_rate_conversion(0.75, units=SI_UNITS)
        self.assertEquals(round(out, 2), 0.73)
        out = pitch_rate_conversion(1.5, units=SI_UNITS)
        self.assertEquals(round(out, 2), 1.46)
        out = pitch_rate_conversion(1.0, units=SI_UNITS)
        self.assertEquals(round(out, 3), 0.976)

    def test_pitch_rate_conversion_imperial(self):
        out = pitch_rate_conversion(0.73, units=IMPERIAL_UNITS)
        self.assertEquals(round(out, 2), 0.75)
        out = pitch_rate_conversion(1.46, units=IMPERIAL_UNITS)
        self.assertEquals(round(out, 2), 1.5)
        out = pitch_rate_conversion(0.976, units=IMPERIAL_UNITS)
        self.assertEquals(round(out, 3), 1.0)


class TestYeastModel(unittest.TestCase):

    def setUp(self):
        self.yeast_model = YeastModel('stir plate')

    def test_yeast_model_raises(self):
        with self.assertRaises(Exception):
            YeastModel('not an allowed method')

    def test_get_growth_rate(self):
        inoculation_rate = 6.17
        with self.assertRaises(NotImplementedError):
            self.yeast_model.get_growth_rate(inoculation_rate)

    def test_get_inoculation_rate(self):
        growth_rate = 4.44
        with self.assertRaises(NotImplementedError):
            self.yeast_model.get_inoculation_rate(growth_rate)


class TestKaiserYeastModel(unittest.TestCase):

    def setUp(self):
        self.yeast_model = KaiserYeastModel()

    def test_get_growth_rate(self):
        inoculation_rate = 1.0
        out = self.yeast_model.get_growth_rate(inoculation_rate)
        self.assertEquals(round(out, 2), 1.4)
        inoculation_rate = 2.5
        out = self.yeast_model.get_growth_rate(inoculation_rate)
        self.assertEquals(round(out, 2), 0.66)
        inoculation_rate = 6.17
        out = self.yeast_model.get_growth_rate(inoculation_rate)
        self.assertEquals(round(out, 2), 0)

    def test_get_inoculation_rate(self):
        growth_rate = 4.44
        out = self.yeast_model.get_inoculation_rate(growth_rate)
        self.assertEquals(round(out, 2), 1.4)
        growth_rate = 1.0
        out = self.yeast_model.get_inoculation_rate(growth_rate)
        self.assertEquals(round(out, 2), 1.99)


class TestWhiteYeastModel(unittest.TestCase):

    def setUp(self):
        self.yeast_model_cls = WhiteYeastModel
        self.yeast_model = self.yeast_model_cls()

    def test_get_growth_rate(self):
        inoculation_rate = 6.17
        out = self.yeast_model.get_growth_rate(inoculation_rate)
        self.assertEquals(round(out, 2), 4.44)

    def test_get_inoculation_rate(self):
        growth_rate = 4.44
        out = self.yeast_model.get_inoculation_rate(growth_rate)
        self.assertEquals(round(out, 2), 6.17)

    def test_get_viability(self):
        out = self.yeast_model.get_viability(0)
        self.assertEqual(out, 1.0)
        out = self.yeast_model.get_viability(30)
        self.assertEqual(out, 0.79)
        out = self.yeast_model.get_viability(143)
        self.assertEqual(out, 0.0)

    def test_get_yeast_pitch_rate(self):
        out = self.yeast_model.get_yeast_pitch_rate()
        expected = {'original_gravity': 1.050,
                    'final_volume': 5.0,
                    'viability': 0.79,
                    'cells': 79.0,
                    'target_pitch_rate': 1.42,
                    'pitch_rate_as_is': 0.32,
                    'pitch_rate_cells': 355.0,
                    'cells_needed': 276.0,
                    'required_growth_rate': 4.49,
                    'units': 'imperial',
                    }
        self.assertEquals(out, expected)

    def test_get_yeast_pitch_rate_two_packs(self):
        out = self.yeast_model.get_yeast_pitch_rate(num_packs=2)
        expected = {'original_gravity': 1.05,
                    'final_volume': 5.0,
                    'viability': 0.79,
                    'cells': 158.0,
                    'target_pitch_rate': 1.42,
                    'pitch_rate_as_is': 0.63,
                    'pitch_rate_cells': 355.0,
                    'cells_needed': 197.0,
                    'required_growth_rate': 2.25,
                    'units': 'imperial',
                    }
        self.assertEquals(out, expected)

    def test_get_yeast_pitch_rate_cells_le_zero(self):
        out = self.yeast_model.get_yeast_pitch_rate(days_since_manufacture=365)
        expected = {'original_gravity': 1.050,
                    'final_volume': 5.0,
                    'viability': 0.0,
                    'cells': 0.0,
                    'target_pitch_rate': 1.42,
                    'pitch_rate_as_is': 0.0,
                    'pitch_rate_cells': 355.0,
                    'cells_needed': 355.0,
                    'required_growth_rate': 0.0,
                    'units': 'imperial',
                    }
        self.assertEquals(out, expected)

    def test_get_yeast_pitch_rate_metric(self):
        self.yeast_model.set_units(SI_UNITS)
        out = self.yeast_model.get_yeast_pitch_rate(
            final_volume=21.0,
            target_pitch_rate=1.5,
            num_packs=2)
        expected = {'original_gravity': 1.050,
                    'final_volume': 21.0,
                    'viability': 0.79,
                    'cells': 158.0,
                    'target_pitch_rate': 1.5,
                    'pitch_rate_as_is': 0.61,
                    'pitch_rate_cells': 390.21,
                    'cells_needed': 232.21,
                    'required_growth_rate': 2.47,
                    'units': 'metric',
                    }
        self.assertEquals(out, expected)

    def test_get_starter_volume(self):
        available_cells = 160.0
        out = self.yeast_model.get_starter_volume(available_cells)
        expected = {'available_cells': 160.0,
                    'starter_volume': 0.53,
                    'original_gravity': 1.036,
                    'dme': 7.23,
                    'inoculation_rate': 80.0,
                    'growth_rate': 0.68,
                    'end_cell_count': 268.15,
                    'units': 'imperial',
                    }
        self.assertEquals(out, expected)

    def test_get_starter_volume_metric_no_agitation(self):
        available_cells = 160.0
        self.yeast_model.set_units(SI_UNITS)
        out = self.yeast_model.get_starter_volume(available_cells,
                                                  starter_volume=2.0)
        expected = {'available_cells': 160.0,
                    'starter_volume': 2.0,
                    'original_gravity': 1.036,
                    'dme': 204.9,
                    'inoculation_rate': 80.0,
                    'growth_rate': 0.68,
                    'end_cell_count': 268.15,
                    'units': 'metric',
                    }
        self.assertEquals(out, expected)

    def test_get_starter_volume_metric_shaking(self):
        available_cells = 160.0
        yeast_model = self.yeast_model_cls('shaking')
        yeast_model.set_units(SI_UNITS)
        out = yeast_model.get_starter_volume(available_cells,
                                             starter_volume=2.0)
        expected = {'available_cells': 160.0,
                    'starter_volume': 2.0,
                    'original_gravity': 1.036,
                    'dme': 204.9,
                    'inoculation_rate': 80.0,
                    'growth_rate': 1.18,
                    'end_cell_count': 348.15,
                    'units': 'metric',
                    }
        self.assertEquals(out, expected)

    def test_get_starter_volume_metric_stir_plate(self):
        available_cells = 160.0
        yeast_model = self.yeast_model_cls('stir plate')
        yeast_model.set_units(SI_UNITS)
        out = yeast_model.get_starter_volume(available_cells,
                                             starter_volume=2.0)
        expected = {'available_cells': 160.0,
                    'starter_volume': 2.0,
                    'original_gravity': 1.036,
                    'dme': 204.9,
                    'inoculation_rate': 80.0,
                    'growth_rate': 1.68,
                    'end_cell_count': 428.15,
                    'units': 'metric',
                    }
        self.assertEquals(out, expected)

    def test_get_resulting_pitch_rate_imperial(self):
        out = self.yeast_model.get_resulting_pitch_rate(268.15)
        self.assertEquals(round(out, 2), 1.49)

    def test_get_resulting_pitch_rate_metric(self):
        self.yeast_model.set_units(SI_UNITS)
        out = self.yeast_model.get_resulting_pitch_rate(268.15,
                                                        final_volume=21.0)
        self.assertEquals(round(out, 2), 1.41)




import argparse
import sys

from brew.utilities.temperature import fahrenheit_to_celsius
from brew.utilities.temperature import celsius_to_fahrenheit


def get_temp_conversion(fahrenheit, celsius):
    """
    Convert temperature between fahrenheit and celsius
    """
    if fahrenheit:
        return round(fahrenheit_to_celsius(fahrenheit), 1)
    elif celsius:
        return round(celsius_to_fahrenheit(celsius), 1)


def get_parser():
    parser = argparse.ArgumentParser(description='Temperature Conversion')
    parser.add_argument('-c', '--celsius', metavar='C', type=float,
                        help='Temperature in Celsius')
    parser.add_argument('-f', '--fahrenheit', metavar='F', type=float,
                        help='Temperature in Fahrenheit')
    return parser


def main():
    parser = get_parser()
    args = parser.parse_args()
    if args.fahrenheit and args.celsius:
        print("Must provide only one of Fahrenheit or Celsius")
        sys.exit(1)
    elif not (args.fahrenheit or args.celsius):
        print("Must provide one of Fahrenheit or Celsius")
        sys.exit(1)
    out = get_temp_conversion(args.fahrenheit, args.celsius)
    print(out)



import unittest

from brew.constants import IMPERIAL_UNITS
from brew.constants import SI_UNITS
from brew.recipes import Recipe
from fixtures import grain_additions
from fixtures import hop_additions
from fixtures import recipe
from fixtures import yeast


class TestRecipe(unittest.TestCase):

    def setUp(self):
        # Define Grains
        self.grain_additions = grain_additions

        # Define Hops
        self.hop_additions = hop_additions

        # Define Yeast
        self.yeast = yeast

        # Define Recipes
        self.recipe = recipe
        self.assertEquals(self.recipe.units, IMPERIAL_UNITS)

    def test_str(self):
        out = str(self.recipe)
        self.assertEquals(out, 'pale ale')

    def test_set_units(self):
        self.assertEquals(self.recipe.units, IMPERIAL_UNITS)
        self.recipe.set_units(SI_UNITS)
        self.assertEquals(self.recipe.units, SI_UNITS)
        self.recipe.set_units(IMPERIAL_UNITS)
        self.assertEquals(self.recipe.units, IMPERIAL_UNITS)

    def test_set_raises(self):
        with self.assertRaises(Exception):
            self.recipe.set_units('bad')

    def test_grains_units_mismatch_raises(self):
        grain_additions = [g.change_units() for g in self.grain_additions]
        with self.assertRaises(Exception):
            Recipe(name='pale ale',
                   grain_additions=grain_additions,
                   hop_additions=self.hop_additions,
                   yeast=self.yeast)

    def test_hops_units_mismatch_raises(self):
        hop_additions = [h.change_units() for h in self.hop_additions]
        with self.assertRaises(Exception):
            Recipe(name='pale ale',
                   grain_additions=self.grain_additions,
                   hop_additions=hop_additions,
                   yeast=self.yeast)



#! /usr/bin/env python

from brew.grains import Grain
from brew.grains import GrainAddition
from brew.hops import Hop
from brew.hops import HopAddition
from brew.recipes import Recipe
from brew.yeasts import Yeast


def main():
    # Define Grains
    pale = Grain('Pale Malt (2 Row) US',
                 short_name='2-row',
                 color=1.8,
                 ppg=37)
    pale_add = GrainAddition(pale,
                             weight=13.96)
    crystal = Grain('Caramel/Crystal Malt - 20L',
                    short_name='C20',
                    color=20.0,
                    ppg=35)
    # OG 32.74
    crystal_add = GrainAddition(crystal,
                                weight=0.78)
    grain_list = [pale, crystal]
    grain_additions = [pale_add, crystal_add]

    # Define Hops
    centennial = Hop('Centennial',
                     percent_alpha_acids=0.14)
    centennial_add = HopAddition(centennial,
                                 weight=0.57,
                                 boil_time=60.0)
    cascade = Hop('Cascade (US)',
                  percent_alpha_acids=0.07)
    cascade_add = HopAddition(cascade,
                              weight=0.76,
                              boil_time=5.0)
    hop_list = [centennial, cascade]
    hop_additions = [centennial_add, cascade_add]

    # Define Yeast
    yeast = Yeast('Wyeast 1056')

    # Define Beer
    beer = Recipe('pale ale',
                  grain_additions=grain_additions,
                  hop_additions=hop_additions,
                  yeast=yeast,
                  percent_brew_house_yield=0.70,  # %
                  start_volume=7.0,  # G
                  final_volume=5.0,  # G
                  )
    print(beer.format())


if __name__ == "__main__":
    main()




import argparse
import sys

from brew.constants import HYDROMETER_ADJUSTMENT_TEMP
from brew.constants import IMPERIAL_UNITS
from brew.constants import SI_UNITS
from brew.utilities.abv import alcohol_by_volume_alternative
from brew.utilities.abv import alcohol_by_volume_standard
from brew.utilities.sugar import hydrometer_adjustment
from brew.utilities.sugar import refractometer_adjustment


def get_abv(og, fg,
            og_temp=HYDROMETER_ADJUSTMENT_TEMP,
            fg_temp=HYDROMETER_ADJUSTMENT_TEMP,
            alternative=False,
            refractometer=False,
            units=IMPERIAL_UNITS,
            verbose=False):
    """
    Get Alcohol by Volume for CLI Utility

    og - Original Specific Gravity
    fg - Final Specific Gravity
    og_temp - Temperature of reading for og
    fg_temp - Temperature of reading for fg
    alternative - Use alternative ABV calculation
    refractometer - Adjust for using a refractometer for readings
    units - Type of units to use in calculations
    verbose - Return verbose information about calculations
    """
    # Gravity is required for calculation
    if not og:
        raise Exception("Original gravity required")
    if not fg:
        raise Exception("Final gravity required")

    # Ensure the gravity units are not mixed up
    if og < fg:
        raise Exception("Original Gravity must be higher than Final Gravity")

    if units not in [IMPERIAL_UNITS, SI_UNITS]:
        raise Exception("Units must be in either {} or {}".format(IMPERIAL_UNITS,  # nopep8
                                                                  SI_UNITS))

    # Adjust the gravity based on temperature
    og = hydrometer_adjustment(og, og_temp, units=units)
    fg = hydrometer_adjustment(fg, fg_temp, units=units)

    # Adjust the final gravity if using a refractometer
    if refractometer:
        fg = refractometer_adjustment(og, fg)

    # Calculate the ABV
    if alternative:
        abv = alcohol_by_volume_alternative(og, fg)
    else:
        abv = alcohol_by_volume_standard(og, fg)

    if verbose:
        out = []
        t_unit = 'F' if units == IMPERIAL_UNITS else 'C'
        out.append("OG     : {:0.3f}".format(og))
        out.append("OG Adj : {:0.3f}".format(og))
        out.append("OG Temp: {:0.2f} {}".format(og_temp, t_unit))
        out.append("FG     : {:0.3f}".format(fg))
        out.append("FG Adj : {:0.3f}".format(fg))
        out.append("FG Temp: {:0.2f} {}".format(fg_temp, t_unit))
        out.append("ABV    : {:0.2f} %".format(abv))
        return '\n'.join(out)
    else:
        return abv


def get_parser():
    parser = argparse.ArgumentParser(description='ABV Calculator')
    parser.add_argument('-o', '--og', metavar='O', type=float,
                        required=True,
                        help='Original Gravity')
    parser.add_argument('-f', '--fg', metavar='F', type=float,
                        required=True,
                        help='Final Gravity')
    parser.add_argument('--og-temp', metavar='T', type=float,
                        default=HYDROMETER_ADJUSTMENT_TEMP,
                        help='Original Gravity Temperature (default: %(default)s)')  # nopep8
    parser.add_argument('--fg-temp', metavar='T', type=float,
                        default=HYDROMETER_ADJUSTMENT_TEMP,
                        help='Final Gravity Temperature (default: %(default)s)')  # nopep8
    parser.add_argument('-a', '--alternative', action='store_true',
                        default=False,
                        help='Use alternative ABV equation')
    parser.add_argument('-r', '--refractometer', action='store_true',
                        default=False,
                        help='Adjust the Final Gravity if using a Refractometer reading')  # nopep8
    parser.add_argument('--units', metavar="U", type=str,
                        default=IMPERIAL_UNITS,
                        help='Units to use (default: %(default)s)')
    parser.add_argument('-v', '--verbose', action='store_true',
                        default=False,
                        help='Verbose Output')
    return parser


def main():
    parser = get_parser()
    args = parser.parse_args()
    try:
        out = get_abv(args.og, args.fg,
                      og_temp=args.og_temp,
                      fg_temp=args.fg_temp,
                      alternative=args.alternative,
                      refractometer=args.refractometer,
                      units=args.units,
                      verbose=args.verbose)
        if args.verbose:
            print(out)
        else:
            print("{:0.2f} %".format(out))
    except Exception as e:
        print(e)
        sys.exit(1)




from ..constants import HYDROMETER_ADJUSTMENT_TEMP
from ..constants import IMPERIAL_UNITS
from ..constants import SI_UNITS
from ..validators import validate_units
from .temperature import celsius_to_fahrenheit


__all__ = [
    'sg_to_gu',
    'gu_to_sg',
    'plato_to_sg',
    'sg_to_plato',
    'brix_to_sg',
    'sg_to_brix',
    'brix_to_plato',
    'plato_to_brix',
    'hydrometer_adjustment',
    'refractometer_adjustment',
]


def sg_to_gu(sg):
    """
    Specific Gravity to Gravity Units
    """
    return (sg - 1) * 1000.0


def gu_to_sg(gu):
    """
    Gravity Units to Specific Gravity
    """
    return 1 + (gu / 1000.0)


def plato_to_sg(deg_plato):
    """
    Degrees Plato to Specific Gravity

    Specific Gravity (S.G.)
    S.G. is the density of a liquid or solid compared to that of water.
    The simple formula for S.G. is:

    S.G. = 1 + 0.004 x Plato

    The more precise calculation of S.G. is:

    S.G. = [(Plato) / (258.6 - (Plato/258.2 x 227.1))] + 1

    Source:
    http://www.learntobrew.com/page/1mdhe/Shopping/Beer_Calculations.html
    """
    return (deg_plato / (258.6 - ((deg_plato / 258.2) * 227.1))) + 1.0


def sg_to_plato(sg):
    """
    Specific Gravity to Degrees Plato

    Degrees Plato is the weight of the extract in a 100gram solution at
    64 degrees Fahrenheit.

    Plato = [(S.G. - 1) x 1000] / 4

    The more precise calculation of Plato is:

    Plato = -616.868 + 1111.14 * sg - 630.272 * sg ** 2 + 135.997 * sg ** 3

    Source:
    http://www.brewersfriend.com/2012/10/31/on-the-relationship-between-plato-and-specific-gravity/
    """
    # return (sg - 1.0) * 1000 / 4
    return ((135.997 * sg - 630.272) * sg + 1111.14) * sg - 616.868


def brix_to_sg(brix):
    """
    Degrees Brix to Specific Gravity

    Source:
    http://www.brewersfriend.com/brix-converter/
    """
    return (brix / (258.6 - ((brix / 258.2) * 227.1))) + 1


def sg_to_brix(sg):
    """
    Specific Gravity to Degrees Brix

    Source:
    http://en.wikipedia.org/wiki/Brix
    http://www.brewersfriend.com/brix-converter/
    """
    if sg > 1.17874:
        raise Exception("Above 40 degBx this function no longer works")
    return (((182.4601 * sg - 775.6821) * sg + 1262.7794) * sg - 669.5622)


def brix_to_plato(brix):
    """
    Degrees Brix to Degrees Plato

    The difference between the degBx and degP as calculated from the respective
    polynomials is:

        degP - degBx = (((-2.81615*sg + 8.79724)*sg - 9.1626)*sg + 3.18213)

    The difference is generally less than +/-0.0005 degBx or degP with the
    exception being for weak solutions.

    https://en.wikipedia.org/wiki/Brix
    """
    return sg_to_plato(brix_to_sg(brix))


def plato_to_brix(plato):
    """
    Degrees Plato to Degrees Brix
    """
    return sg_to_brix(plato_to_sg(plato))


def hydrometer_adjustment(sg, temp, units=IMPERIAL_UNITS):
    """
    Adjust the Hydrometer if the temperature deviates from 59degF.

    http://hbd.org/brewery/library/HydromCorr0992.html
    Correction(@59F) = 1.313454 - 0.132674*T + 2.057793e-3*T**2 - 2.627634e-6*T**3
        where T is in degrees F.

    Sources:
    http://www.brewersfriend.com/hydrometer-temp/
    http://www.topdownbrew.com/SGCorrection.html
    http://hbd.org/brewery/library/HydromCorr0992.html
    http://www.primetab.com/formulas.html
    """  # nopep8
    validate_units(units)
    if units == SI_UNITS:
        if temp < 0.0 or 100.0 < temp:
            raise Exception("Correction does not work outside temps 0 - 100C")
        temp = celsius_to_fahrenheit(temp)
    elif units == IMPERIAL_UNITS:
        if temp < 0.0 or 212.0 < temp:
            raise Exception("Correction does not work outside temps 0 - 212F")

    if temp == HYDROMETER_ADJUSTMENT_TEMP:
        return sg

    correction = (1.313454 - 0.132674 * (temp ** 1) +
                  (2.057793 * 10 ** -3) * (temp ** 2) -
                  (2.627634 * 10 ** -6) * (temp ** 3))
    return sg + (correction * 0.001)


def refractometer_adjustment(og, fg):
    """
    Adjust the Refractometer for the presence of alcohol.

    NOTE: This calculation assumes using Brix or Plato, so the input will be
    converted from SG to Plato and then converted back.

    Returns: Final Gravity

    Sources:
    http://seanterrill.com/2011/04/07/refractometer-fg-results/
    """
    og_brix = sg_to_brix(og)
    fg_brix = sg_to_brix(fg)

    new_fg = (1.0000 -
              0.0044993 * og_brix + 0.011774 * fg_brix +
              0.00027581 * (og_brix ** 2) - 0.0012717 * (fg_brix ** 2) -
              0.0000072800 * (og_brix ** 3) + 0.000063293 * (fg_brix ** 3))
    return brix_to_sg(new_fg)




from ..constants import GAL_PER_LITER
from ..constants import IMPERIAL_UNITS
from ..constants import POUND_PER_KG
from ..constants import SI_UNITS
from ..validators import validate_units


__all__ = [
    'srm_to_ebc',
    'ebc_to_srm',
    'calculate_mcu',
    'calculate_srm_mosher',
    'calculate_srm_daniels',
    'calculate_srm_daniels_power',
    'calculate_srm_noonan_power',
    'calculate_srm_morey_hybrid',
    'calculate_srm_morey',
    'calculate_srm',
    'lovibond_to_srm',
    'srm_to_lovibond',
    'srm_to_a430',
    'ebc_to_a430',
]


def srm_to_ebc(srm):
    """
    Convert SRM to EBC Color
    """
    return srm * 1.97


def ebc_to_srm(ebc):
    """
    Convert EBC to SRM Color
    """
    return ebc / 1.97


def calculate_mcu(grain_weight, beer_color, final_volume,
                  units=IMPERIAL_UNITS):
    """
    Calculate MCU from Grain

    grain_weight - in lbs or kg
    beer_color - in deg Lovibond
    final_volume - in gal or liters

    http://beersmith.com/blog/2008/04/29/beer-color-understanding-srm-lovibond-and-ebc/
    """  # nopep8
    validate_units(units)
    if units == SI_UNITS:
        grain_weight = grain_weight * POUND_PER_KG
        final_volume = final_volume * GAL_PER_LITER

    mcu = grain_weight * beer_color / final_volume
    return mcu


def calculate_srm_mosher(mcu):
    """
    Mosher Equation for SRM

    grain_weight - in lbs or kg
    beer_color - in deg Lovibond
    final_volume - in gal or liters
    """  # nopep8
    if mcu < 7.0:
        raise Exception("Mosher equation does not work for MCU < 7.0")
    srm = (mcu * 0.3) + 4.7
    return srm


def calculate_srm_daniels(mcu):
    """
    Daniels Equation for SRM

    grain_weight - in lbs or kg
    beer_color - in deg Lovibond
    final_volume - in gal or liters
    """  # nopep8
    # if mcu < 11.0:
    #     raise Exception("Daniels equation does not work for MCU < 11.0")
    srm = (mcu * 0.2) + 8.4
    return srm


def calculate_srm_daniels_power(mcu):
    """
    Daniels Power Equation for SRM based on work by Druey

    grain_weight - in lbs or kg
    beer_color - in deg Lovibond
    final_volume - in gal or liters
    """  # nopep8
    srm = 1.73 * (mcu ** 0.64) - 0.27
    if srm > 50.0:
        raise Exception("Daniels Power equation does not work above SRM 50")
    return srm


def calculate_srm_noonan_power(mcu):
    """
    Noonan Power Equation for SRM based on work by Druey

    grain_weight - in lbs or kg
    beer_color - in deg Lovibond
    final_volume - in gal or liters
    """  # nopep8
    srm = 15.03 * (mcu ** 0.27) - 15.53
    if srm > 50.0:
        raise Exception("Noonan Power equation does not work above SRM 50")
    return srm


def calculate_srm_morey_hybrid(mcu):
    """
    A hybrid approach used by Morey for SRM.

    Assumptions:

    1. SRM is approximately equal to MCU for values from 0 to 10.
    2. Homebrew is generally darker than commercial beer.
    3. Base on the previous qualitative postulate, I assumed that Ray Daniels'
       predicted relationship exists for beers with color greater than 10.
    4. Since Mosher's equation predicts darker color than Daniels' model for
       values of MCU greater than 37, I assumed that Mosher's approximation
       governed beer color for all values more than 37 MCUs.
    5. Difference in color for beers greater than 40 SRM are essentially
       impossible to detect visually; therefore, I limited the analysis to SRM
       of 50 and less.

    http://babblehomebrewers.com/attachments/article/61/beercolor.pdf
    """
    if 0 < mcu < 10:
        return mcu
    elif 10 <= mcu < 37:
        return calculate_srm_daniels(mcu)
    elif 37 <= mcu < 50:
        return calculate_srm_mosher(mcu)
    else:
        raise Exception("Morey Hybrid does not work above SRM 50")


def calculate_srm_morey(mcu):
    """
    Morey Equation for SRM

    http://www.morebeer.com/brewingtechniques/beerslaw/morey.html

    grain_weight - in lbs or kg
    beer_color - in deg Lovibond
    final_volume - in gal or liters

    http://beersmith.com/blog/2008/04/29/beer-color-understanding-srm-lovibond-and-ebc/
    """  # nopep8
    srm = 1.4922 * (mcu ** 0.6859)
    if srm > 50.0:
        raise Exception("Morey equation does not work above SRM 50")
    return srm


def calculate_srm(mcu):
    """
    General srm calculation uses the Morey Power Equation
    """
    return calculate_srm_morey(mcu)


def lovibond_to_srm(lovibond):
    """
    Convert deg Lovibond to SRM
    https://en.wikipedia.org/wiki/Standard_Reference_Method
    """
    return 1.3546 * lovibond - 0.76


def srm_to_lovibond(srm):
    """
    Convert SRM to deg Lovibond
    https://en.wikipedia.org/wiki/Standard_Reference_Method
    """
    return (srm + 0.76) / 1.3546


def srm_to_a430(srm, dilution=1.0):
    """
    Get attenuation at A430 from SRM and dilution
    https://en.wikipedia.org/wiki/Standard_Reference_Method
    """
    return srm / (12.7 * dilution)


def ebc_to_a430(ebc, dilution=1.0):
    """
    Get attenuation at A430 from EBC and dilution
    https://en.wikipedia.org/wiki/Standard_Reference_Method
    """
    return ebc / (25 * dilution)



from .constants import GRAIN_TYPE_LIST
from .constants import HOP_TYPE_LIST
from .constants import IMPERIAL_UNITS
from .constants import SI_UNITS


__all__ = [
    'validate_grain_type',
    'validate_hop_type',
    'validate_percentage',
    'validate_units',
    'validate_required_fields',
    'validate_optional_fields',
]


def validate_grain_type(grain_type):
    if grain_type in GRAIN_TYPE_LIST:
        return grain_type
    raise Exception("Unkown grain type '{}', must use {}".format(
        grain_type, ', '.join(GRAIN_TYPE_LIST)))


def validate_hop_type(hop_type):
    if hop_type in HOP_TYPE_LIST:
        return hop_type
    raise Exception("Unkown hop type '{}', must use {}".format(
        hop_type, ', '.join(HOP_TYPE_LIST)))


def validate_percentage(percent):
    if 0.0 <= percent <= 1.0:
        return percent
    raise Exception("Percentage values should be in decimal format")


def validate_units(units):
    if units in [IMPERIAL_UNITS, SI_UNITS]:
        return units
    raise Exception("Unkown units '{}', must use {} or {}".format(
        units, IMPERIAL_UNITS, SI_UNITS))


def validate_required_fields(data, required_fields):
    """
    Validate fields which are required as part of the data.

    data: a python dict
    required_fields: a list of tuples where the first element is a string with
                     a value that should be a key found in the data dict and
                     where the second element is a python type or list/tuple of
                     python types to check the field against.
    """
    for field, field_type in required_fields:
        if field not in data:
            raise Exception("Required field '{}' missing from recipe".format(
                field))
        if not isinstance(data[field], field_type):
            raise Exception("Required field '{}' is not of type '{}'".format(
                field, field_type))


def validate_optional_fields(data, optional_fields, data_field='data'):
    """
    Validate fields which are optional as part of the data.

    data: a python dict
    optional_fields: a list of tuples where the first element is a string with
                     a value that should be a key found in the data dict and
                     where the second element is a python type or list/tuple of
                     python types to check the field against.
    """
    # If no optional data field present then return
    if data_field not in data:
        return
    for field, field_type in optional_fields:
        if field in data[data_field]:
            # With optional fields only check the type as they are overrides
            # and not all overrides need to be present
            if not isinstance(data[data_field][field], field_type):
                raise Exception("Optional field '{}' in '{}' is not of type '{}'".format(  # nopep8
                    field, data_field, field_type))




from ..constants import FC_DIFF_TWO_ROW
from ..constants import LITERS_OF_WORT_AT_SG
from ..constants import MOISTURE_CORRECTION
from ..constants import MOISTURE_FINISHED_MALT
from ..constants import PPG_TO_HWE_CONVERSION
from ..constants import SUCROSE_PLATO
from ..constants import SUCROSE_PPG
from ..validators import validate_percentage
from .sugar import gu_to_sg


__all__ = [
    'dry_to_liquid_malt_weight',
    'liquid_to_dry_malt_weight',
    'grain_to_liquid_malt_weight',
    'liquid_malt_to_grain_weight',
    'dry_malt_to_grain_weight',
    'grain_to_dry_malt_weight',
    'specialty_grain_to_liquid_malt_weight',
    'liquid_malt_to_specialty_grain_weight',
    'fine_grind_to_coarse_grind',
    'coarse_grind_to_fine_grind',
    'dry_basis_to_as_is_basis',
    'as_is_basis_to_dry_basis',
    'sg_from_dry_basis',
    'plato_from_dry_basis',
    'basis_to_hwe',
    'hwe_to_basis',
    'ppg_to_hwe',
    'hwe_to_ppg',
]


def dry_to_liquid_malt_weight(malt):
    """
    DME to LME Weight

    Source: http://www.weekendbrewer.com/brewingformulas.htm
    """
    return malt * 1.25


def liquid_to_dry_malt_weight(malt):
    """
    LME to DME Weight

    Source: http://www.weekendbrewer.com/brewingformulas.htm
    """
    return malt / 1.25


def grain_to_liquid_malt_weight(grain):
    """
    Grain to LME Weight

    Source: http://www.weekendbrewer.com/brewingformulas.htm
    """
    return grain * 0.75


def liquid_malt_to_grain_weight(malt):
    """
    LME to Grain Weight
    """
    return malt / 0.75


def dry_malt_to_grain_weight(malt):
    """
    DME to Grain Weight
    """
    return malt * 5.0 / 3.0


def grain_to_dry_malt_weight(malt):
    """
    Grain to DME Weight
    """
    return malt * 3.0 / 5.0


def specialty_grain_to_liquid_malt_weight(grain):
    """
    Specialty Grain to LME Weight

    Source: http://www.weekendbrewer.com/brewingformulas.htm
    """
    return grain * 0.89


def liquid_malt_to_specialty_grain_weight(malt):
    """
    LME to Specialty Grain Weight
    """
    return malt / 0.89


def fine_grind_to_coarse_grind(fine_grind, fc_diff=FC_DIFF_TWO_ROW):
    """
    Fine Grind to Coarse Grind Percentage

    fine_grind is a percentage from the malt bill
    fc_diff is the F/C difference percentage from the malt bill
    """
    validate_percentage(fine_grind)
    validate_percentage(fc_diff)
    return fine_grind - fc_diff


def coarse_grind_to_fine_grind(coarse_grind, fc_diff=FC_DIFF_TWO_ROW):
    """
    Coarse Grind to Fine Grind Percentage

    coarse_grind is a percentage from the malt bill
    fc_diff is the F/C difference percentage from the malt bill
    """
    validate_percentage(coarse_grind)
    validate_percentage(fc_diff)
    return coarse_grind + fc_diff


def dry_basis_to_as_is_basis(dry_basis,
                             moisture_content=MOISTURE_FINISHED_MALT):
    """
    Dry Basis to As-Is Basis Percentage

    dry_basis is a percentage from the malt bill in decimal form
    moisture_content is a percentage of moisture content in finished malt
      in decimal form
    """
    validate_percentage(dry_basis)
    validate_percentage(moisture_content)
    return dry_basis * (1.0 - moisture_content)


def as_is_basis_to_dry_basis(as_is,
                             moisture_content=MOISTURE_FINISHED_MALT):
    """
    As-Is Basis to Dry Basis Percentage

    asi_is is a percentage from the malt bill in decimal form
    moisture_content is a percentage of moisture content in finished malt
      in decimal form
    """
    validate_percentage(as_is)
    validate_percentage(moisture_content)
    return as_is / (1.0 - moisture_content)


def sg_from_dry_basis(dbcg,
                      moisture_content=MOISTURE_FINISHED_MALT,
                      moisture_correction=MOISTURE_CORRECTION,
                      brew_house_efficiency=0.90):
    """
    Specific Gravity from Dry Basis Percentage

    dbcg is Dry Basis Coarse Grain in decimal form
    moisture_content is a percentage of moisture content in finished malt
      in decimal form
    moisture_correction is a percentage correction in decimal form
    brew_house_efficiency is the efficiency in decimal form

    Returns: Specific Gravity available from Malt
    """
    validate_percentage(dbcg)
    validate_percentage(moisture_content)
    validate_percentage(moisture_correction)
    validate_percentage(brew_house_efficiency)
    gu = ((dbcg - moisture_content - moisture_correction) *
          brew_house_efficiency * SUCROSE_PPG)
    return gu_to_sg(gu)


def plato_from_dry_basis(dbcg,
                         moisture_content=MOISTURE_FINISHED_MALT,
                         moisture_correction=MOISTURE_CORRECTION,
                         brew_house_efficiency=0.90):
    """
    Degrees Plato from Dry Basis Percentage

    dbcg is Dry Basis Coars Grain in decimal form
    moisture_content is a percentage of moisture content in finished malt
      in decimal form
    moisture_correction is a percentage correction in decimal form
    brew_house_efficiency is the efficiency in decimal form

    Returns: Degrees Plato available from Malt
    """
    validate_percentage(dbcg)
    validate_percentage(moisture_content)
    validate_percentage(moisture_correction)
    validate_percentage(brew_house_efficiency)
    return ((dbcg - moisture_content - moisture_correction) *
            brew_house_efficiency * SUCROSE_PLATO)


def basis_to_hwe(basis_percentage):
    """
    Basis Percentage to Hot Water Extract

    basis_percentage in decimal form

    Return Hot Water Extract as Ldeg/kg, dry basis

    Ldeg/kg means how many litres of wort with a specific gravity of 1.001 you
    could produce from a kilogram of the fermentable

    For example, if you had a kilogram of sucrose, you could make up 386 litres
    of wort with a specific gravity of 1.001.
    """
    validate_percentage(basis_percentage)
    return basis_percentage * LITERS_OF_WORT_AT_SG


def hwe_to_basis(hwe):
    """
    Hot Water Extract to Basis Percentage
    """
    return hwe / LITERS_OF_WORT_AT_SG


def ppg_to_hwe(ppg):
    """
    Points Per Gallon to Hot Water Extract
    """
    return ppg * PPG_TO_HWE_CONVERSION


def hwe_to_ppg(hwe):
    """
    Hot Water Extract to Points Per Gallon
    """
    return hwe / PPG_TO_HWE_CONVERSION



import json
import textwrap

from .constants import HOP_TYPE_PELLET
from .constants import HOPS_CONSTANT_IMPERIAL
from .constants import HOPS_CONSTANT_SI
from .constants import IMPERIAL_TYPES
from .constants import IMPERIAL_UNITS
from .constants import MG_PER_OZ
from .constants import OZ_PER_MG
from .constants import SI_TYPES
from .constants import SI_UNITS
from .utilities.hops import HopsUtilizationGlennTinseth
from .validators import validate_hop_type
from .validators import validate_optional_fields
from .validators import validate_percentage
from .validators import validate_required_fields
from .validators import validate_units


__all__ = ['Hop', 'HopAddition']


class Hop(object):
    """
    A representation of a type of Hop.
    """

    def __init__(self, name,
                 percent_alpha_acids=None):
        """
        Percent Alpha Acids - The percent alpha acids in the hop.
        """
        self.name = name
        self.percent_alpha_acids = validate_percentage(percent_alpha_acids)

    def __str__(self):
        return "{0}, alpha {1}%".format(self.name.capitalize(),
                                        self.percent_alpha_acids)

    def __repr__(self):
        out = "{0}('{1}'".format(type(self).__name__, self.name)
        if self.percent_alpha_acids:
            out = "{0}, percent_alpha_acids={1}".format(
                out, self.percent_alpha_acids)
        out = "{0})".format(out)
        return out

    def to_dict(self):
        return {'name': self.name,
                'percent_alpha_acids': round(self.percent_alpha_acids, 2),
                }

    def to_json(self):
        return json.dumps(self.to_dict(), sort_keys=True)

    def format(self):
        msg = textwrap.dedent("""\
                {name} Hops
                -----------------------------------
                Alpha Acids:  {percent_alpha_acids} %""".format(
            **self.to_dict()))
        return msg


class HopAddition(object):
    """
    A representation of the Hop as added to a Recipe.
    """

    def __init__(self, hop,
                 weight=None,
                 boil_time=None,
                 hop_type=HOP_TYPE_PELLET,
                 utilization_cls=HopsUtilizationGlennTinseth,
                 utilization_cls_kwargs=None,
                 units=IMPERIAL_UNITS):
        """
        Weight - The weight of the hop to add
        Boil Time - The time the hop should be boiled for measured backwards
            from the end of the boil.
        Hop Type - The type of hop being used
        Utilization Class - The utilization calculation class to use
        Utilization Class Kwargs - The keyword arguments to initialize the
            Utilization Class with
        """
        self.hop = hop
        self.weight = weight
        self.boil_time = boil_time
        self.hop_type = validate_hop_type(hop_type)
        self.utilization_cls_kwargs = utilization_cls_kwargs or {}
        self.utilization_cls = utilization_cls(self, **self.utilization_cls_kwargs)  # nopep8

        # Manage units
        self.set_units(units)

    def set_units(self, units):
        self.units = validate_units(units)
        if self.units == IMPERIAL_UNITS:
            self.types = IMPERIAL_TYPES
        elif self.units == SI_UNITS:
            self.types = SI_TYPES

    def change_units(self):
        """
        Change units from one type to the other return new instance
        """
        if self.units == IMPERIAL_UNITS:
            weight = self.weight * MG_PER_OZ
            units = SI_UNITS
        elif self.units == SI_UNITS:
            weight = self.weight * OZ_PER_MG
            units = IMPERIAL_UNITS
        return HopAddition(self.hop,
                           weight=weight,
                           boil_time=self.boil_time,
                           utilization_cls_kwargs={'units': units},
                           units=units)

    def __str__(self):
        return "{hop}, {weight} {weight_small}, {boil_time} min, {hop_type}".format(  # nopep8
                hop=self.hop,
                weight=self.weight,
                boil_time=self.boil_time,
                hop_type=self.hop_type,
                **self.types)

    def __repr__(self):
        out = "{0}({1}".format(type(self).__name__, repr(self.hop))
        if self.weight:
            out = "{0}, weight={1}".format(out, self.weight)
        if self.boil_time:
            out = "{0}, boil_time={1}".format(out, self.boil_time)
        if self.hop_type:
            out = "{0}, hop_type='{1}'".format(out, self.hop_type)
        if self.utilization_cls:
            out = "{0}, utilization_cls={1}".format(out, type(self.utilization_cls).__name__)  # nopep8
        if self.utilization_cls_kwargs:
            out = "{0}, utilization_cls_kwargs={1}".format(out, self.utilization_cls_kwargs)  # nopep8
        if self.units:
            out = "{0}, units='{1}'".format(out, self.units)
        out = "{0})".format(out)
        return out

    def to_dict(self):
        hop_data = self.hop.to_dict()
        return {'name': hop_data.pop('name'),
                'data': hop_data,
                'weight': round(self.weight, 2),
                'boil_time': round(self.boil_time, 1),
                'hop_type': self.hop_type,
                'utilization_cls': str(self.utilization_cls),
                'utilization_cls_kwargs': self.utilization_cls_kwargs,
                'units': self.units,
                }

    def to_json(self):
        return json.dumps(self.to_dict(), sort_keys=True)

    @classmethod
    def validate(cls, hop_data):
        required_fields = [('name', str),
                           ('weight', float),
                           ('boil_time', float),
                           ]
        optional_fields = [('percent_alpha_acids', float),
                           ('hop_type', str),
                           ('units', str),
                           ]
        validate_required_fields(hop_data, required_fields)
        validate_optional_fields(hop_data, optional_fields)

    def format(self):
        kwargs = {}
        kwargs.update(self.to_dict())
        kwargs.update(self.types)
        msg = textwrap.dedent("""\
                {name} Addition
                -----------------------------------
                Hop Type:     {hop_type}
                AA %:         {data[percent_alpha_acids]:0.2f} %
                Weight:       {weight:0.2f} {weight_small}
                Boil Time:    {boil_time:0.1f} min""".format(
            **kwargs))
        return msg

    def get_ibus(self, sg, final_volume):
        return self.utilization_cls.get_ibus(sg, final_volume)

    def get_alpha_acid_units(self):
        """
        Alpha Acid Units

        Defined as ounces of hops * alpha acids
        """
        alpha_acids = self.hop.percent_alpha_acids * 100
        if self.units == IMPERIAL_UNITS:
            return self.weight * alpha_acids
        elif self.units == SI_UNITS:
            return self.weight * OZ_PER_MG * alpha_acids

    def get_hops_weight(self, sg, target_ibu, final_volume,
                        percent_contribution):
        """
        Weight of Hops
        IBUs or International Bittering Units measures a bitterness unit for hops.
        IBUs are the measurement in parts per million (ppm) of iso-alpha acids
        in the beer.   For example, an IPA with 75 IBUs has 75 milligrams of
        isomerized alpha acids per liter. The equation used to calculate the
        weight of hops for the boil is as follows.

        Ounces hops = (IBU Target)(galbeer)(IBU%) / (%a-acid)(%Utilization)(7489)

        The IBU target equals the total bitterness for the beer.  (e.g. an IPA
        may have an IBU target of 75 IBUs)  The percent IBU is equal to the
        percent of IBUs from each hop addition.  You may wish for your first hop
        addition to contribute 95% of the total IBUs.  This would make your
        IBU% 95%.  The %a-acid is the amount of alpha acid in the hops and can
        be found on the hop packaging.  The % Utilization is a measurement of
        the percentage of alpha acid units that will isomerize in the boil.
        The following chart outlines the typical utilizations and hop boil times.

        60 min = 30% utilization
        30 min = 15%
        5   min = 2.5%

        The 7489 is a conversion factor and used to cancel the units in the
        equation, converting oz/gallon to mg/l. For the hops equation, the
        units for the % must be expressed in decimal form.  (e.g. 10%= .10)

        Source:
        - http://www.learntobrew.com/page/1mdhe/Shopping/Beer_Calculations.html
        # nopep8
        """
        validate_percentage(percent_contribution)

        hops_constant = HOPS_CONSTANT_IMPERIAL
        if self.units == SI_UNITS:
            hops_constant = HOPS_CONSTANT_SI

        utilization = self.utilization_cls.get_percent_utilization(
            sg, self.boil_time)

        num = (target_ibu * final_volume)
        den = (utilization * self.hop.percent_alpha_acids * hops_constant)
        return num / den * percent_contribution



import unittest

from brew.utilities.abv import alcohol_by_volume_alternative
from brew.utilities.abv import alcohol_by_volume_standard


class TestABVUtilities(unittest.TestCase):

    def test_alcohol_by_volume_alternative(self):
        abv = alcohol_by_volume_alternative(1.057, 1.013)
        self.assertEquals(round(abv, 2), 5.95)

    def test_alcohol_by_volume_standard(self):
        abv = alcohol_by_volume_standard(1.057, 1.013)
        self.assertEquals(round(abv, 2), 5.78)




import unittest

from brew.cli.temp import get_temp_conversion


class TestCliTemp(unittest.TestCase):

    def setUp(self):
        self.c = 15.6
        self.f = 60.1

    def test_get_sugar_conversion_fahrenheit(self):
        out = get_temp_conversion(self.f, None)
        self.assertEquals(out, self.c)

    def test_get_sugar_conversion_celsius(self):
        out = get_temp_conversion(None, self.c)
        self.assertEquals(out, self.f)



import unittest

from brew.parsers import JSONDataLoader


class TestJSONDataLoader(unittest.TestCase):

    def setUp(self):
        self.parser = JSONDataLoader('./')

    def test_format_name(self):
        name_list = [('pale malt 2-row us', 'pale_malt_2_row_us'),
                     ('caramel crystal malt 20l', 'caramel_crystal_malt_20l'),
                     ('centennial', 'centennial'),
                     ('cascade us', 'cascade_us'),
                     ('Wyeast 1056', 'wyeast_1056'),
                     ]
        for name, expected in name_list:
            out = self.parser.format_name(name)
            self.assertEquals(out, expected)



from ..constants import ALCOHOL_SPECIFIC_GRAVITY


__all__ = [
    'alcohol_by_volume_standard',
    'alcohol_by_volume_alternative',
    'alcohol_by_weight',
]


def alcohol_by_volume_standard(og, fg):
    """
    Alcohol by Volume Standard Calculation

    Most brewing sites use this basic formula:

    ABV = (og - fg) * 131.25

    This equation was created before the computer age.  It is easy to do by
    hand, and over time became the accepted formula for home brewers!

    Variations on this equation which report within tenths of each other
    come from The Joy of Homebrewing Method by Charlie Papazian, Bee Lee's
    Method, Beer Advocate Method. Some variations use 131 instead of 131.25.
    The resulting difference is pretty minor.

    Source:
    - http://www.brewersfriend.com/2011/06/16/alcohol-by-volume-calculator-updated/
    """  # nopep8
    return (og - fg) * 131.25


def alcohol_by_volume_alternative(og, fg):
    """
    Alcohol by Volume Alternative Calculation

    Alternate Formula:

    A more complex equation which attempts to provide greater accuracy at higher gravities is:

    ABV =(76.08 * (og - fg) / (1.775 - og)) * (fg / 0.794)

    The alternate equation reports a higher ABV for higher gravity beers.
    This equation is just a different take on it. Scientists rarely agree
    when it comes to equations. There will probably be another equation for
    ABV down the road.

    The complex formula, and variations on it come from
    Ritchie Products Ltd, (Zymurgy, Summer 1995, vol. 18, no. 2)
    -Michael L. Hall's article Brew by the Numbers: Add Up What's in Your
    Beer, and Designing Great Beers by Daniels.

    Source:
    - http://www.brewersfriend.com/2011/06/16/alcohol-by-volume-calculator-updated/
    """  # nopep8
    return (76.08 * (og - fg) / (1.775 - og)) * (fg / 0.794)


def alcohol_by_weight(abv):
    """
    Alcohol by Weight from ABV
    """
    return abv * ALCOHOL_SPECIFIC_GRAVITY



import unittest

from brew.constants import KG_PER_POUND
from brew.constants import LITER_PER_GAL
from brew.constants import SI_UNITS
from brew.utilities.color import calculate_mcu
from brew.utilities.color import calculate_srm
from brew.utilities.color import calculate_srm_daniels
from brew.utilities.color import calculate_srm_daniels_power
from brew.utilities.color import calculate_srm_morey
from brew.utilities.color import calculate_srm_morey_hybrid
from brew.utilities.color import calculate_srm_mosher
from brew.utilities.color import calculate_srm_noonan_power
from brew.utilities.color import ebc_to_a430
from brew.utilities.color import ebc_to_srm
from brew.utilities.color import lovibond_to_srm
from brew.utilities.color import srm_to_a430
from brew.utilities.color import srm_to_ebc
from brew.utilities.color import srm_to_lovibond


class TestColorUtilities(unittest.TestCase):

    def test_srm_to_ebc(self):
        ebc = srm_to_ebc(3.0)
        self.assertEquals(round(ebc, 2), 5.91)

    def test_ebc_to_srm(self):
        srm = ebc_to_srm(5.91)
        self.assertEquals(round(srm, 2), 3.0)

    def test_calculate_mcu(self):
        weight = 1.0  # lbs
        color = 30.0  # degL
        vol = 5.5  # gal
        mcu = calculate_mcu(weight, color, vol)
        self.assertEquals(round(mcu, 2), 5.45)

    def test_calculate_srm_imperial(self):
        weight = 1.0  # lbs
        color = 30.0  # degL
        vol = 5.5  # gal
        mcu = calculate_mcu(weight, color, vol)
        srm = calculate_srm(mcu)
        self.assertEquals(round(srm, 2), 4.78)
        ebc = srm_to_ebc(srm)
        self.assertEquals(round(ebc, 2), 9.41)

    def test_calculate_srm_metric(self):
        weight = 1.0 * KG_PER_POUND  # kg
        color = 30.0  # degL
        vol = 5.5 * LITER_PER_GAL  # liter
        self.assertEquals(round(weight, 2), 0.45)
        self.assertEquals(round(vol, 2), 20.82)

        mcu = calculate_mcu(weight, color, vol, units=SI_UNITS)
        srm = calculate_srm(mcu)
        self.assertEquals(round(srm, 2), 4.78)
        ebc = srm_to_ebc(srm)
        self.assertEquals(round(ebc, 2), 9.41)

    def test_calculate_srm_all(self):
        weight = 3.0  # lbs
        color = 30.0  # degL
        vol = 5.5  # gal
        mcu = calculate_mcu(weight, color, vol)
        srm = calculate_srm_mosher(mcu)
        self.assertEquals(round(srm, 2), 9.61)
        srm = calculate_srm_daniels(mcu)
        self.assertEquals(round(srm, 2), 11.67)
        srm = calculate_srm_daniels_power(mcu)
        self.assertEquals(round(srm, 2), 10.08)
        srm = calculate_srm_noonan_power(mcu)
        self.assertEquals(round(srm, 2), 16.44)
        srm = calculate_srm_morey_hybrid(mcu)
        self.assertEquals(round(srm, 2), 11.67)
        srm = calculate_srm_morey(mcu)
        self.assertEquals(round(srm, 2), 10.15)

    def test_calculate_srm_morey_hybrid(self):
        mcu = calculate_mcu(1.0, 30.0, 5.5)
        self.assertTrue(0 < mcu < 10)
        srm = calculate_srm_morey_hybrid(mcu)
        self.assertEquals(round(srm, 2), 5.45)

        mcu = calculate_mcu(3.0, 30.0, 5.5)
        self.assertTrue(10 <= mcu < 37)
        srm = calculate_srm_morey_hybrid(mcu)
        self.assertEquals(round(srm, 2), 11.67)

        mcu = calculate_mcu(8.0, 30.0, 5.5)
        self.assertTrue(37 <= mcu < 50)
        srm = calculate_srm_morey_hybrid(mcu)
        self.assertEquals(round(srm, 2), 17.79)

        mcu = calculate_mcu(100.0, 30.0, 5.5)
        self.assertTrue(50 < mcu)
        with self.assertRaises(Exception):
            calculate_srm_morey_hybrid(mcu)

    def test_calculate_srm_mosher_raises(self):
        weight = 1.0  # lbs
        color = 30.0  # degL
        vol = 5.5  # gal
        mcu = calculate_mcu(weight, color, vol)
        with self.assertRaises(Exception):
            calculate_srm_mosher(mcu)

    # def test_calculate_srm_daniels_raises(self):
    #     weight = 1.0  # lbs
    #     color = 30.0  # degL
    #     vol = 5.5  # gal
    #     mcu = calculate_mcu(weight, color, vol)
    #     with self.assertRaises(Exception):
    #         calculate_srm_daniels(mcu)

    def test_calculate_srm_daniels_power_raises(self):
        weight = 100.0  # lbs
        color = 30.0  # degL
        vol = 5.5  # gal
        mcu = calculate_mcu(weight, color, vol)
        with self.assertRaises(Exception):
            calculate_srm_daniels_power(mcu)

    def test_calculate_srm_noonan_power_raises(self):
        weight = 100.0  # lbs
        color = 30.0  # degL
        vol = 5.5  # gal
        mcu = calculate_mcu(weight, color, vol)
        with self.assertRaises(Exception):
            calculate_srm_noonan_power(mcu)

    def test_calculate_srm_morey_raises(self):
        weight = 1.0  # lbs
        color = 1000.0  # degL
        vol = 1.0  # gal
        mcu = calculate_mcu(weight, color, vol)
        with self.assertRaises(Exception):
            calculate_srm_morey(mcu)

    def test_calculate_srm_raises(self):
        weight = 1.0  # lbs
        color = 1000.0  # degL
        vol = 1.0  # gal
        mcu = calculate_mcu(weight, color, vol)
        with self.assertRaises(Exception):
            calculate_srm(mcu)

    def test_lovibond_to_srm(self):
        out = lovibond_to_srm(30)
        self.assertEquals(round(out, 2), 39.88)

    def test_srm_to_lovibond(self):
        out = srm_to_lovibond(40)
        self.assertEquals(round(out, 2), 30.09)

    def test_srm_to_a430(self):
        out = srm_to_a430(30)
        self.assertEquals(round(out, 2), 2.36)

    def test_ebc_to_a430(self):
        out = ebc_to_a430(60)
        self.assertEquals(round(out, 2), 2.4)



#! /usr/bin/env python

"""
Brewing Classic Styles: 80 Winning Recipes Anyone Can Brew
by Jamil Zainasheff and John J. Palmer

Raison de Saison

Used by permission of Brewers Publications (2007). All rights reserved.

You can purchase the book here:

- http://www.brewerspublications.com/books/brewing-classic-styles-80-winning-recipes-anyone-can-brew/

Original Stats:

OG:      1.060 (14.8P)
FG:      1.008 ( 2.0P)
ADF:     86%
IBU:     27
Color:   5 SRM (10 EBC)
Alcohol: 6.9% ABV (5.4% ABW)
Boil:    60 min
Pre-Boil Volume:  7 Gal (26.5L)
Pre-Boil Gravity: 1.051 (12.7P)
"""  # nopep8

import os

from brew.parsers import JSONDataLoader
from brew.parsers import parse_recipe


def main():

    recipe = {
        'name': "Raison de Saison (Extract)",
        'start_volume': 7.0,
        'final_volume': 6.0,
        'grains': [
            {'name': 'Pilsner Liquid Extract',
             'data': {
                 'color': 2.3,
                 'ppg': 37,
             },
             'weight': 7.7,
             'grain_type': 'lme'},
            {'name': 'Cane Beet Sugar',
             'weight': 1.0,
             'grain_type': 'dme'},
            {'name': 'Wheat Liquid Extract',
             'data': {
                 'color': 4.0,
                 'ppg': 37,
             },
             'weight': 0.75,
             'grain_type': 'lme'},
            {'name': 'munich liquid malt extract',
             'data': {
                 'color': 9.0,
                 'ppg': 37,
             },
             'weight': 0.5,
             'grain_type': 'lme'},
            {'name': 'Caramunich Malt',
             'data': {
                 'color': 60.0,
             },
             'weight': 0.125,
             'grain_type': 'specialty'},
        ],
        'hops': [
            {'name': 'Hallertau US',
             'data': {
                 'percent_alpha_acids': 0.05,
             },
             'weight': 1.7,
             'boil_time': 60.0},
            {'name': 'Hallertau US',
             'data': {
                 'percent_alpha_acids': 0.05,
             },
             'weight': 0.75,
             'boil_time': 0.0},
        ],
        'yeast': {
            'name': 'Wyeast 3724',
            'data': {
                'percent_attenuation': 0.86,
            },
        },
        'data': {
            'percent_brew_house_yield': 0.70,
            'units': 'imperial',
        },
    }

    data_dir = os.path.abspath(os.path.join(os.getcwd(), 'data/'))
    loader = JSONDataLoader(data_dir)
    beer = parse_recipe(recipe, loader)
    print(beer.format())


if __name__ == "__main__":
    main()




"""
Yeast Pitch Calculator

This replicates other calculators found here:
    - http://www.brewersfriend.com/yeast-pitch-rate-and-starter-calculator/

"""  # nopep8

import argparse
import textwrap
import sys

from brew.constants import GAL_PER_LITER
from brew.constants import IMPERIAL_UNITS
from brew.constants import SI_UNITS
from brew.utilities.yeast import KaiserYeastModel
from brew.utilities.yeast import WhiteYeastModel


def get_yeast_pitch_calculation(args):
    if args.units not in [IMPERIAL_UNITS, SI_UNITS]:
        print("Units must be in either {} or {}".format(IMPERIAL_UNITS,
                                                        SI_UNITS))
        sys.exit(1)

    model_cls = None
    if args.model == 'white':
        model_cls = WhiteYeastModel
    elif args.model == 'kaiser':
        model_cls = KaiserYeastModel
    else:
        print("Unknown Yeast Growth Model '{}', must be 'white' or 'kaiser'".format(args.model))  # nopep8
        sys.exit(1)

    try:
        model = model_cls(args.method, units=args.units)
    except Exception as e:
        print(e)
        sys.exit(1)
    data = model.get_yeast_pitch_rate(original_gravity=args.og,
                                      final_volume=args.fv,
                                      target_pitch_rate=args.target_pitch_rate,
                                      yeast_type=args.type,
                                      cells_per_pack=args.cells,
                                      num_packs=args.num,
                                      days_since_manufacture=args.days)
    cells_needed = data['cells_needed']
    data.update(model.types)
    msg = textwrap.dedent("""\
            Yeast Pitch Calculator
            -----------------------------------
            Original Gravity      {original_gravity:0.3f}
            Final Volume          {final_volume:0.2f} {volume}
            Target Pitch Rate     {target_pitch_rate}
            Viability             {viability} %
            Cells                 {cells:0.0f} B
            Pitch Rate As-Is      {pitch_rate_as_is:0.2f}
            Pitch Rate Cells      {pitch_rate_cells:0.0f} B
            Cells Needed          {cells_needed} B
            Required Growth Rate  {required_growth_rate}
            Units                 {units}""".format(**data))
    print(msg)

    if data['cells'] <= 0:
        print("\nNo cells available for further calculation")
        sys.exit(1)

    data = model.get_starter_volume(available_cells=data['cells'],
                                    starter_volume=args.sv,
                                    original_gravity=args.sg)
    end_cell_count = data['end_cell_count']
    data.update(model.types)
    msg = textwrap.dedent("""\

            Starter Volume
            -----------------------------------
            Available Cells       {available_cells:0.0f}
            Starter Volume        {starter_volume:0.2f} {volume}
            Original Gravity      {original_gravity:0.3f}
            DME Required          {dme:0.2f} {weight_small}
            Inoculation Rate      {inoculation_rate}
            Growth Rate           {growth_rate}
            End Cell Count        {end_cell_count} B
            Units                 {units}""".format(**data))
    print(msg)

    if cells_needed < end_cell_count:
        print("\nStarter has enough cells")
    else:
        print("\nStarter DOES NOT HAVE enough cells")

    resulting_pitch_rate = model.get_resulting_pitch_rate(
        starter_cell_count=end_cell_count,
        original_gravity=args.sg,
        final_volume=args.fv)
    data.update(model.types)
    msg = textwrap.dedent("""\

            Resulting Pitch Rate
            -----------------------------------
            Starter Cell Count    {starter_cell_count} B
            Original Gravity      {original_gravity}
            Final Volume          {final_volume} {volume}
            Pitch Rate            {resulting_pitch_rate}""".format(
        starter_cell_count=end_cell_count,
        original_gravity=args.sg,
        final_volume=args.fv,
        resulting_pitch_rate=resulting_pitch_rate,
        **model.types))
    print(msg)


def get_parser():
    parser = argparse.ArgumentParser(description='Yeast Pitch Calculator')
    parser.add_argument('--og', metavar='O', type=float,
                        default=1.05,
                        help='Wort Original Gravity (default: %(default)s)')
    parser.add_argument('--fv', metavar='V', type=float,
                        default=5.0,
                        help='Wort Final Volume (default: %(default)s)')
    parser.add_argument('--sg', metavar='O', type=float,
                        default=1.036,
                        help='Starter Original Gravity (default: %(default)s)')
    parser.add_argument('--sv', metavar='V', type=float,
                        default=2.0 * GAL_PER_LITER,
                        help='Starter Volume (default: %(default)s)')
    parser.add_argument('--target-pitch-rate', metavar='T', type=float,
                        default=1.42,
                        help='Target Pitch Rate (default: %(default)s)')
    parser.add_argument('--type', metavar='T', type=str,
                        default='liquid',
                        help='Yeast Type (default: %(default)s)')
    parser.add_argument('-c', '--cells', metavar='C', type=int,
                        default=100,
                        help='Number of cells per container in Billions')
    parser.add_argument('-n', '--num', metavar='N', type=int,
                        default=1,
                        help='Number of containers (default: %(default)s)')
    parser.add_argument('-d', '--days', metavar='D', type=int,
                        default=0,
                        help='Number of days since yeast manufacture (default: %(default)s)')  # nopep8
    parser.add_argument('--model', metavar='M', type=str,
                        default='white',
                        help='Model of yeast growth, white or kaiser (default: %(default)s)')  # nopep8
    parser.add_argument('--method', metavar='M', type=str,
                        default='stir plate',
                        help='Method of growth (default: %(default)s)')  # nopep8
    parser.add_argument('--units', metavar="U", type=str,
                        default=IMPERIAL_UNITS,
                        help='Units to use (default: %(default)s)')
    return parser


def main():
    parser = get_parser()
    args = parser.parse_args()
    get_yeast_pitch_calculation(args)



#! /usr/bin/python

import json
import string
import textwrap

from .constants import GRAIN_TYPE_DME
from .constants import GRAIN_TYPE_LME
from .constants import GAL_PER_LITER
from .constants import HOP_TYPE_PELLET
from .constants import HOP_UTILIZATION_SCALE_PELLET
from .constants import IMPERIAL_TYPES
from .constants import IMPERIAL_UNITS
from .constants import LITER_PER_GAL
from .constants import SI_TYPES
from .constants import SI_UNITS
from .constants import WATER_WEIGHT_IMPERIAL
from .constants import WATER_WEIGHT_SI
from .utilities.abv import alcohol_by_volume_alternative
from .utilities.abv import alcohol_by_volume_standard
from .utilities.abv import alcohol_by_weight
from .utilities.color import calculate_mcu
from .utilities.color import calculate_srm
from .utilities.color import calculate_srm_daniels
from .utilities.color import calculate_srm_morey
from .utilities.color import calculate_srm_mosher
from .utilities.color import srm_to_ebc
from .utilities.sugar import gu_to_sg
from .utilities.sugar import sg_to_plato
from .validators import validate_optional_fields
from .validators import validate_percentage
from .validators import validate_required_fields
from .validators import validate_units


__all__ = ['Recipe']


class Recipe(object):
    """
    A representation of a Recipe that can be brewed to make beer.
    """
    grain_lookup = {}
    hop_lookup = {}

    def __init__(self, name,
                 grain_additions=None,
                 hop_additions=None,
                 yeast=None,
                 percent_brew_house_yield=0.70,
                 start_volume=7.0,
                 final_volume=5.0,
                 units=IMPERIAL_UNITS):
        self.name = name
        self.grain_additions = grain_additions
        self.hop_additions = hop_additions
        self.yeast = yeast

        self.percent_brew_house_yield = validate_percentage(percent_brew_house_yield)  # nopep8 %
        self.start_volume = start_volume  # G
        self.final_volume = final_volume  # G

        # Manage units
        self.set_units(units)

        # For each grain and hop:
        # 1. Add to lookup
        # 2. Ensure all units are the same
        for grain_add in self.grain_additions:
            self.grain_lookup[grain_add.grain.name] = grain_add
            if grain_add.units != self.units:
                raise Exception("Grain addition units must be in '{}' not '{}'".format(  # nopep8
                    self.units, grain_add.units))
        for hop_add in self.hop_additions:
            # The same hops may be used several times, so we must distinguish
            hop_key = '{}_{}'.format(hop_add.hop.name, hop_add.boil_time)
            self.hop_lookup[hop_key] = hop_add
            if hop_add.units != self.units:
                raise Exception("Hop addition units must be in '{}' not '{}'".format(  # nopep8
                    self.units, hop_add.units))

    def __str__(self):
        return self.name

    def set_units(self, units):
        self.units = validate_units(units)
        if self.units == IMPERIAL_UNITS:
            self.types = IMPERIAL_TYPES
        elif self.units == SI_UNITS:
            self.types = SI_TYPES

    def change_units(self):
        """
        Change units from one type to the other return new instance
        """
        if self.units == IMPERIAL_UNITS:
            start_volume = self.start_volume * LITER_PER_GAL
            final_volume = self.final_volume * LITER_PER_GAL
            units = SI_UNITS
        elif self.units == SI_UNITS:
            start_volume = self.start_volume * GAL_PER_LITER
            final_volume = self.final_volume * GAL_PER_LITER
            units = IMPERIAL_UNITS
        return Recipe(self.name,
                      grain_additions=[ga.change_units() for ga in
                                       self.grain_additions],
                      hop_additions=[ha.change_units() for ha in
                                     self.hop_additions],
                      yeast=self.yeast,
                      percent_brew_house_yield=self.percent_brew_house_yield,
                      start_volume=start_volume,
                      final_volume=final_volume,
                      units=units)

    def get_total_points(self):
        # Pick the attribute based on units
        if self.units == IMPERIAL_UNITS:
            attr = 'ppg'
        if self.units == SI_UNITS:
            attr = 'hwe'

        total_points = 0
        for grain_add in self.grain_additions:
            # DME and LME are 100% efficient in disolving in water
            # Cereal extraction depends on brew house yield
            efficiency = self.percent_brew_house_yield
            if grain_add.grain_type in [GRAIN_TYPE_DME, GRAIN_TYPE_LME]:
                efficiency = 1.0
            total_points += getattr(grain_add.grain, attr) * grain_add.weight * efficiency  # nopep8
        return total_points

    def get_original_gravity_units(self):
        return self.get_total_points() / self.final_volume

    def get_original_gravity(self):
        return gu_to_sg(self.get_original_gravity_units())

    def get_boil_gravity_units(self):
        return self.get_total_points() / self.start_volume

    def get_boil_gravity(self):
        return gu_to_sg(self.get_boil_gravity_units())

    def get_final_gravity_units(self):
        return self.get_original_gravity_units() * (1.0 - self.yeast.percent_attenuation)  # nopep8

    def get_final_gravity(self):
        return gu_to_sg(self.get_final_gravity_units())

    def get_degrees_plato(self):
        return sg_to_plato(self.get_boil_gravity())

    def get_brew_house_yield(self, plato_actual, vol_actual):
        """
        Brew House Yield (BHY)
        Brew house yield is a measurement that tells the efficiency of the
        brewing.  The actual degrees Plato from the brew and the actual gallons
        collected out of the kettle are needed to calculate the BHY.

        BHY  =  [(Pactual)(galactual)(BHYtarget)] / [(Ptarget)(galtarget)]
        """
        num = plato_actual * vol_actual * self.percent_brew_house_yield
        den = self.get_degrees_plato() * self.final_volume
        return num / den

    def get_extract_weight(self):
        """
        Weight of Extract
        The weight of extract is the amount of malt extract present in the
        wort.

        Lbs extract = (density of water) * (gal of wort) * (SG) * (P/100)

        The weight of one gallon of water in the above formula is 8.32 lbs/gal

        To find the weight of a gallon of wort, multiply the specific gravity
        of the wort by the density of water.

        Plato is a percentage of sugars by weight.  So 10 Plato means solution
        is 10% sugars.  In this equation we convert the degrees plato to a
        decimal number between 0.0 and 1.0 by dividing it by 100.  This is
        multiplied by the  weight of a gallon of wort.
        """
        water_density = WATER_WEIGHT_IMPERIAL
        if self.units == SI_UNITS:
            water_density = WATER_WEIGHT_SI
        return (water_density * self.final_volume * self.get_boil_gravity() *
                (self.get_degrees_plato() / 100.0))

    def get_percent_malt_bill(self, grain_add):
        """
        Percent malt bill is how much extract each grain addition adds to the
        recipe. To ensure different additions are measured equally each is
        converted to dry weight.
        """
        return self.get_grain_add_dry_weight(grain_add) / self.get_total_dry_weight()  # nopep8

    def get_grain_add_dry_weight(self, grain_add):
        """
        When converting Grain to DME its important to remember
        that you can't get 100% efficiency from grains.  Multiplying by
        the brew house yield will decrease the size of the DME
        accordingly.
        """
        if grain_add.grain_type in [GRAIN_TYPE_DME, GRAIN_TYPE_LME]:
            return grain_add.get_dry_weight()
        else:
            return grain_add.get_dry_weight() * self.percent_brew_house_yield  # nopep8

    def get_total_dry_weight(self):
        weights = []
        for grain_add in self.grain_additions:
            weights.append(self.get_grain_add_dry_weight(grain_add))
        return sum(weights)

    def get_grain_add_cereal_weight(self, grain_add):
        """
        When converting DME or LME to grain its important to remember
        that you can't get 100% efficiency from grains.  Dividing by
        the brew house yield will increase the size of the grain
        accordingly.
        """
        if grain_add.grain_type in [GRAIN_TYPE_DME, GRAIN_TYPE_LME]:
            return grain_add.get_cereal_weight() / self.percent_brew_house_yield  # nopep8
        else:
            return grain_add.get_cereal_weight()

    def get_total_grain_weight(self):
        weights = []
        for grain_add in self.grain_additions:
            weights.append(self.get_grain_add_cereal_weight(grain_add))
        return sum(weights)

    def get_percent_ibus(self, hop_add):
        """
        Get the percentage the hops contributes to total ibus
        """
        bg = self.get_boil_gravity()
        fv = self.final_volume
        return hop_add.get_ibus(bg, fv) / self.get_total_ibu()

    def get_total_ibu(self):
        """
        Convenience method to get total IBU for the recipe
        """
        bg = self.get_boil_gravity()
        fv = self.final_volume
        return sum([hop_add.get_ibus(bg, fv)
                   for hop_add in self.hop_additions])

    def get_bu_to_gu(self):
        """
        Returns ratio of Bitterness Units to Original Gravity Units
        """
        return self.get_total_ibu() / self.get_boil_gravity_units()

    @classmethod
    def get_strike_temp(cls, mash_temp, malt_temp, liquor_to_grist_ratio):
        """
        Strike Water Temp
        As you know when you are mashing, your strike water has to be warmer
        than the target mash temperature because the cool malt will cool the
        temperature of the water.  To correctly calculate the temperature of
        the strike water, use the following formula.

        Strike Temp =  [((0.4)(T mash-T malt)) / L:G] +  T mash
        """
        return (((0.4 * (mash_temp - malt_temp)) /
                liquor_to_grist_ratio) + mash_temp)

    def get_mash_water_volume(self, liquor_to_grist_ratio):
        """
        Mash Water Volume
        To calculate the mash water volume you will need to know your liquor to
        grist ratio.  The term liquor refers to the mash water and grist refers
        to the milled malt.  We need to calculate the appropriate amount of
        water to allow for enzyme action and starch conversion take place.

        gallons H2O =  (Lbs malt)(L:G)(1gallon H2O) / 8.32 pounds water
        """
        water_weight = WATER_WEIGHT_IMPERIAL
        if self.units == SI_UNITS:
            water_weight = WATER_WEIGHT_SI
        return (self.get_total_dry_weight() * liquor_to_grist_ratio /
                water_weight)

    def get_wort_color_mcu(self, grain_add):
        """
        Calculation of Wort and Beer Color

        Color of Wort = S [(% extract)(L of malt)(P wort / 8P reference)]

        Source:
        http://beersmith.com/blog/2008/04/29/beer-color-understanding-srm-lovibond-and-ebc/
        http://brewwiki.com/index.php/Estimating_Color
        """  # nopep8
        weight = self.get_grain_add_cereal_weight(grain_add)
        return calculate_mcu(weight,
                             grain_add.grain.color,
                             self.final_volume,
                             units=self.units)

    def get_wort_color(self, grain_add):
        mcu = self.get_wort_color_mcu(grain_add)
        return calculate_srm(mcu)

    def get_total_wort_color(self):
        """
        Convenience method to get total wort color
        """
        mcu = sum([self.get_wort_color_mcu(ga) for ga in self.grain_additions])
        return calculate_srm(mcu)

    def get_total_wort_color_map(self):
        """
        Convenience method to get total wort color
        """
        mcu = sum([self.get_wort_color_mcu(ga) for ga in self.grain_additions])
        srm_morey = calculate_srm_morey(mcu)
        srm_daniels = calculate_srm_daniels(mcu)
        srm_mosher = calculate_srm_mosher(mcu)

        return {
            'srm': {
                'morey': round(srm_morey, 1),
                'daniels': round(srm_daniels, 1),
                'mosher': round(srm_mosher, 1),
            },
            'ebc': {
                'morey': round(srm_to_ebc(srm_morey), 1),
                'daniels': round(srm_to_ebc(srm_daniels), 1),
                'mosher': round(srm_to_ebc(srm_mosher), 1),
            },
        }

    def to_dict(self):
        og = self.get_original_gravity()
        bg = self.get_boil_gravity()
        fg = self.get_final_gravity()
        abv_standard = alcohol_by_volume_standard(og, fg)
        abv_alternative = alcohol_by_volume_alternative(og, fg)
        recipe_dict = {
            'name': string.capwords(self.name),
            'start_volume': round(self.start_volume, 2),
            'final_volume': round(self.final_volume, 2),
            'data': {
                'percent_brew_house_yield': round(self.percent_brew_house_yield, 2),  # nopep8
                'original_gravity': round(og, 3),
                'boil_gravity': round(bg, 3),
                'final_gravity': round(fg, 3),
                'abv_standard': round(abv_standard, 2),
                'abv_alternative': round(abv_alternative, 2),
                'abw_standard': round(alcohol_by_weight(abv_standard), 2),
                'abw_alternative': round(alcohol_by_weight(abv_alternative), 2),  # nopep8
                'total_wort_color_map': self.get_total_wort_color_map(),
                'total_ibu': round(self.get_total_ibu(), 1),
                'bu_to_gu': round(self.get_bu_to_gu(), 1),
                'units': self.units,
            },
            'grains': [],
            'hops': [],
            'yeast': {},
        }

        for grain_add in self.grain_additions:
            grain = grain_add.to_dict()
            wort_color_srm = self.get_wort_color(grain_add)
            wort_color_ebc = srm_to_ebc(wort_color_srm)
            working_yield = round(grain_add.grain.get_working_yield(self.percent_brew_house_yield), 2)  # nopep8
            percent_malt_bill = round(self.get_percent_malt_bill(grain_add), 2)
            grain['data'].update({
                'working_yield': working_yield,
                'percent_malt_bill': percent_malt_bill,
                'wort_color_srm': round(wort_color_srm, 1),
                'wort_color_ebc': round(wort_color_ebc, 1),
            })
            recipe_dict['grains'].append(grain)

        for hop_add in self.hop_additions:
            hop = hop_add.to_dict()

            ibus = hop_add.get_ibus(bg, self.final_volume)

            utilization = hop_add.utilization_cls.get_percent_utilization(
                bg, hop_add.boil_time)
            # Utilization is 10% higher for pellet vs whole/plug
            if hop_add.hop_type == HOP_TYPE_PELLET:
                utilization *= HOP_UTILIZATION_SCALE_PELLET

            hop['data'].update({
                'ibus': round(ibus, 1),
                'utilization': round(utilization, 2),
            })
            recipe_dict['hops'].append(hop)

        recipe_dict['yeast'] = self.yeast.to_dict()

        return recipe_dict

    def to_json(self):
        return json.dumps(self.to_dict(), sort_keys=True)

    @classmethod
    def validate(cls, recipe):
        required_fields = [('name', str),
                           ('start_volume', (int, float)),
                           ('final_volume', (int, float)),
                           ('grains', (list, tuple)),
                           ('hops', (list, tuple)),
                           ('yeast', dict),
                           ]
        optional_fields = [('percent_brew_house_yield', float),
                           ('units', str),
                           ]
        validate_required_fields(recipe, required_fields)
        validate_optional_fields(recipe, optional_fields)

    def format(self):
        recipe_data = self.to_dict()
        kwargs = {}
        kwargs.update(recipe_data)
        kwargs.update(self.types)

        msg = ""
        msg += textwrap.dedent("""\
            {name}
            ===================================

            Brew House Yield:   {data[percent_brew_house_yield]:0.2f} %
            Start Volume:       {start_volume:0.1f}
            Final Volume:       {final_volume:0.1f}

            Original Gravity:   {data[original_gravity]:0.3f}
            Boil Gravity:       {data[boil_gravity]:0.3f}
            Final Gravity:      {data[final_gravity]:0.3f}

            ABV / ABW Standard: {data[abv_standard]:0.2f} % / {data[abw_standard]:0.2f} %
            ABV / ABW Alt:      {data[abv_alternative]:0.2f} % / {data[abw_alternative]:0.2f} %

            IBU:                {data[total_ibu]:0.1f} ibu
            BU/GU:              {data[bu_to_gu]:0.1f}

            Morey   (SRM/EBC):  {data[total_wort_color_map][srm][morey]:0.1f} degL / {data[total_wort_color_map][ebc][morey]:0.1f}
            Daneils (SRM/EBC):  {data[total_wort_color_map][srm][daniels]:0.1f} degL / {data[total_wort_color_map][ebc][daniels]:0.1f}
            Mosher  (SRM/EBC):  {data[total_wort_color_map][srm][mosher]:0.1f} degL / {data[total_wort_color_map][ebc][mosher]:0.1f}

            """.format(**kwargs))  # nopep8

        msg += textwrap.dedent("""\
            Grains
            ===================================

            """)

        for grain_data in recipe_data['grains']:
            grain_kwargs = {}
            grain_kwargs.update(grain_data)
            grain_kwargs.update(self.types)

            grain_name = grain_data['name']
            grain_add = self.grain_lookup[grain_name]
            if grain_add.units != self.units:
                grain_add = grain_add.change_units()

            msg += grain_add.format()
            msg += textwrap.dedent("""\

                    Percent Malt Bill: {data[percent_malt_bill]:0.2f} %
                    Working Yield:     {data[working_yield]:0.2f} %
                    SRM/EBC:           {data[wort_color_srm]:0.1f} degL / {data[wort_color_ebc]:0.1f}

                    """.format(**grain_kwargs))  # nopep8

        msg += textwrap.dedent("""\
            Hops
            ===================================

            """)

        for hop_data in recipe_data['hops']:
            hop_kwargs = {}
            hop_kwargs.update(hop_data)
            hop_kwargs.update(self.types)

            hop_key = '{}_{}'.format(hop_data['name'],
                                     hop_data['boil_time'])
            hop = self.hop_lookup[hop_key]
            if hop.units != self.units:
                hop = hop.change_units()

            msg += hop.format()
            msg += textwrap.dedent("""\

                    IBUs:         {data[ibus]:0.1f}
                    Utilization:  {data[utilization]:0.2f} %

                    """.format(**hop_kwargs))

        msg += textwrap.dedent("""\
            Yeast
            ===================================

            """)

        msg += self.yeast.format()
        return msg



import json
import string
import textwrap

from .constants import GRAIN_TYPE_CEREAL
from .constants import GRAIN_TYPE_DME
from .constants import GRAIN_TYPE_LME
from .constants import GRAIN_TYPE_SPECIALTY
from .constants import IMPERIAL_TYPES
from .constants import IMPERIAL_UNITS
from .constants import KG_PER_POUND
from .constants import POUND_PER_KG
from .constants import SI_TYPES
from .constants import SI_UNITS
from .utilities.malt import dry_to_liquid_malt_weight
from .utilities.malt import dry_malt_to_grain_weight
from .utilities.malt import grain_to_dry_malt_weight
from .utilities.malt import grain_to_liquid_malt_weight
from .utilities.malt import hwe_to_basis
from .utilities.malt import hwe_to_ppg
from .utilities.malt import liquid_malt_to_grain_weight
from .utilities.malt import liquid_to_dry_malt_weight
from .utilities.malt import specialty_grain_to_liquid_malt_weight
from .utilities.malt import ppg_to_hwe
from .validators import validate_optional_fields
from .validators import validate_percentage
from .validators import validate_required_fields
from .validators import validate_units


__all__ = ['Grain', 'GrainAddition']


class Grain(object):
    """
    A representation of a type of grain.
    """

    def __init__(self, name,
                 short_name=None,
                 color=None,
                 ppg=None,
                 hwe=None):
        """
        Color - The color of the grain in SRM
        PPG - The potential points points per gallon.
        Hot Water Extract - The international unit for the total soluble
            extract of a malt, based on specific gravity. HWE is measured as
            liter*degrees per kilogram, and is equivalent to
            points/pound/gallon (PPG) when you apply metric conversion factors
            for volume and weight. The combined conversion factor is
            8.3454 X PPG = HWE.
        """
        self.name = name
        self.short_name = short_name or name
        self.color = float(color)
        if ppg and hwe:
            raise Exception("Cannot provide both ppg and hwe")
        if ppg:
            self.ppg = float(ppg)
            self.hwe = ppg_to_hwe(self.ppg)
        elif hwe:
            self.hwe = float(hwe)
            self.ppg = hwe_to_ppg(self.hwe)

    def __str__(self):
        return string.capwords(self.name)

    def __repr__(self):
        out = "{0}('{1}'".format(type(self).__name__, self.name)
        if self.short_name:
            out = "{0}, short_name='{1}'".format(out, self.short_name)
        if self.color:
            out = "{0}, color={1}".format(out, self.color)
        if self.hwe:
            out = "{0}, hwe={1}".format(out,
                                                      round(self.hwe, 2))  # nopep8
        out = "{0})".format(out)
        return out

    def to_dict(self):
        return {'name': self.name,
                'short_name': self.short_name,
                'color': round(self.color, 1),
                'ppg': round(self.ppg, 2),
                'hwe': round(self.hwe, 2),
                }

    def to_json(self):
        return json.dumps(self.to_dict(), sort_keys=True)

    def format(self):
        msg = textwrap.dedent("""\
                {name} Grain
                -----------------------------------
                Color:             {color} degL
                PPG:               {ppg:0.2f}
                Hot Water Extract: {hwe:0.2f}""".format(
            **self.to_dict()))
        return msg

    def get_working_yield(self, percent_brew_house_yield):
        """
        Working Yield
        Working Yield is the product of the Hot Water Extract multiplied by the
        Brew House Yield.  This product will provide the percent of extract
        collected from the malt.

        WY =    (HWE as-is)(BHY)
        """
        validate_percentage(percent_brew_house_yield)
        return (hwe_to_basis(self.hwe) *
                percent_brew_house_yield)


class GrainAddition(object):
    """
    A representation of the grain as added to a Recipe.
    """

    def __init__(self, grain,
                 weight=None,
                 grain_type=GRAIN_TYPE_CEREAL,
                 units=IMPERIAL_UNITS):
        """
        Weight - The weight of the grain to add
        Grain Type - The type of grain being used
        """
        self.grain = grain
        self.weight = weight
        self.grain_type = grain_type

        # Manage units
        self.set_units(units)

    def set_units(self, units):
        self.units = validate_units(units)
        if self.units == IMPERIAL_UNITS:
            self.types = IMPERIAL_TYPES
        elif self.units == SI_UNITS:
            self.types = SI_TYPES

    def change_units(self):
        """
        Change units from one type to the other return new instance
        """
        if self.units == IMPERIAL_UNITS:
            weight = self.weight * KG_PER_POUND
            units = SI_UNITS
        elif self.units == SI_UNITS:
            weight = self.weight * POUND_PER_KG
            units = IMPERIAL_UNITS
        return GrainAddition(self.grain,
                             weight=weight,
                             units=units)

    def __str__(self):
        return "{grain}, weight {weight} {weight_large}".format(
            grain=self.grain,
            weight=self.weight,
            **self.types)

    def __repr__(self):
        out = "{0}({1}".format(type(self).__name__, repr(self.grain))
        if self.weight:
            out = "{0}, weight={1}".format(out, self.weight)
        if self.grain_type:
            out = "{0}, grain_type='{1}'".format(out, self.grain_type)
        out = "{0})".format(out)
        return out

    def get_cereal_weight(self):
        """
        Get the weight of the addition in cereal weight
        """
        if self.grain_type == GRAIN_TYPE_CEREAL:
            return self.weight
        elif self.grain_type == GRAIN_TYPE_DME:
            return dry_malt_to_grain_weight(self.weight)
        elif self.grain_type == GRAIN_TYPE_LME:
            return liquid_malt_to_grain_weight(self.weight)
        elif self.grain_type == GRAIN_TYPE_SPECIALTY:
            return self.weight

    def get_lme_weight(self):
        """
        Get the weight of the addition in Liquid Malt Extract weight
        """
        if self.grain_type == GRAIN_TYPE_CEREAL:
            return grain_to_liquid_malt_weight(self.weight)
        elif self.grain_type == GRAIN_TYPE_DME:
            return dry_to_liquid_malt_weight(self.weight)
        elif self.grain_type == GRAIN_TYPE_LME:
            return self.weight
        elif self.grain_type == GRAIN_TYPE_SPECIALTY:
            return specialty_grain_to_liquid_malt_weight(self.weight)

    def get_dry_weight(self):
        """
        Get the weight of the addition in Dry Malt Extract weight
        """
        if self.grain_type == GRAIN_TYPE_CEREAL:
            return grain_to_dry_malt_weight(self.weight)
        elif self.grain_type == GRAIN_TYPE_DME:
            return self.weight
        elif self.grain_type == GRAIN_TYPE_LME:
            return liquid_to_dry_malt_weight(self.weight)
        elif self.grain_type == GRAIN_TYPE_SPECIALTY:
            lme = specialty_grain_to_liquid_malt_weight(self.weight)
            return liquid_to_dry_malt_weight(lme)

    def get_weight_map(self):
        return {
            'grain_weight': round(self.get_cereal_weight(), 2),
            'lme_weight': round(self.get_lme_weight(), 2),
            'dry_weight': round(self.get_dry_weight(), 2),
        }

    def to_dict(self):
        grain_data = self.grain.to_dict()
        return {'name': grain_data.pop('name'),
                'data': grain_data,
                'weight': round(self.weight, 2),
                'grain_type': self.grain_type,
                'units': self.units,
                }

    def to_json(self):
        return json.dumps(self.to_dict(), sort_keys=True)

    @classmethod
    def validate(cls, grain_data):
        required_fields = [('name', str),
                           ('weight', float),
                           ]
        optional_fields = [('color', (int, float)),
                           ('ppg', (int, float)),
                           ('hwe', (int, float)),
                           ('grain_type', str),
                           ('units', str),
                           ]
        validate_required_fields(grain_data, required_fields)
        validate_optional_fields(grain_data, optional_fields)

    def format(self):
        kwargs = {}
        kwargs.update(self.to_dict())
        kwargs.update(self.types)
        msg = textwrap.dedent("""\
                {name} Addition
                -----------------------------------
                Grain Type:        {grain_type}
                Weight:            {weight:0.2f} {weight_large}""".format(
            **kwargs))
        return msg



import unittest

from brew.cli.sugar import get_sugar_conversion


class TestCliSugar(unittest.TestCase):

    def setUp(self):
        self.brix = 22.0
        self.plato = 22.0
        self.sg = 1.092

    def test_get_sugar_conversion_brix_to_brix(self):
        out = get_sugar_conversion(self.brix, None, None, 'b')
        self.assertEquals(round(out, 1), self.brix)

    def test_get_sugar_conversion_brix_to_plato(self):
        out = get_sugar_conversion(self.brix, None, None, 'p')
        self.assertEquals(round(out, 1), self.plato)

    def test_get_sugar_conversion_brix_to_sg(self):
        out = get_sugar_conversion(self.brix, None, None, 's')
        self.assertEquals(round(out, 3), self.sg)

    def test_get_sugar_conversion_plato_to_brix(self):
        out = get_sugar_conversion(None, self.plato, None, 'b')
        self.assertEquals(round(out, 1), self.brix)

    def test_get_sugar_conversion_plato_to_plato(self):
        out = get_sugar_conversion(None, self.plato, None, 'p')
        self.assertEquals(round(out, 1), self.plato)

    def test_get_sugar_conversion_plato_to_sg(self):
        out = get_sugar_conversion(None, self.plato, None, 's')
        self.assertEquals(round(out, 3), self.sg)

    def test_get_sugar_conversion_sg_to_brix(self):
        out = get_sugar_conversion(None, None, self.sg, 'b')
        self.assertEquals(round(out, 1), self.brix)

    def test_get_sugar_conversion_sg_to_plato(self):
        out = get_sugar_conversion(None, None, self.sg, 'p')
        self.assertEquals(round(out, 1), self.plato)

    def test_get_sugar_conversion_sg_to_sg(self):
        out = get_sugar_conversion(None, None, self.sg, 's')
        self.assertEquals(round(out, 3), self.sg)

    def test_get_sugar_conversion_all_brix(self):
        out = get_sugar_conversion(self.brix, None, None, None)
        expected = 'SG\tPlato\tBrix\n1.092\t22.0\t22.0'
        self.assertEquals(out, expected)

    def test_get_sugar_conversion_all_plato(self):
        out = get_sugar_conversion(None, self.plato, None, None)
        expected = 'SG\tPlato\tBrix\n1.092\t22.0\t22.0'
        self.assertEquals(out, expected)

    def test_get_sugar_conversion_all_sg(self):
        out = get_sugar_conversion(None, None, self.sg, None)
        expected = 'SG\tPlato\tBrix\n1.092\t22.0\t22.0'
        self.assertEquals(out, expected)



import textwrap
import unittest

from brew.constants import LITER_PER_GAL
from brew.constants import IMPERIAL_UNITS
from brew.constants import SI_UNITS
from brew.hops import HopAddition
from brew.utilities.sugar import plato_to_sg
from fixtures import cascade
from fixtures import centennial


class TestHops(unittest.TestCase):

    def setUp(self):

        # Define Hops
        self.hop = centennial

    def test_str(self):
        out = str(self.hop)
        self.assertEquals(out, 'Centennial, alpha 0.14%')

    def test_repr(self):
        out = repr(self.hop)
        self.assertEquals(out, "Hop('centennial', percent_alpha_acids=0.14)")

    def test_to_dict(self):
        out = self.hop.to_dict()
        expected = {'name': 'centennial',
                    'percent_alpha_acids': 0.14,
                    }
        self.assertEquals(out, expected)

    def test_to_json(self):
        out = self.hop.to_json()
        expected = '{"name": "centennial", "percent_alpha_acids": 0.14}'
        self.assertEquals(out, expected)

    def test_format(self):
        out = self.hop.format()
        msg = textwrap.dedent("""\
                centennial Hops
                -----------------------------------
                Alpha Acids:  0.14 %""")
        self.assertEquals(out, msg)


class TestHopAdditions(unittest.TestCase):

    def setUp(self):
        self.hop1 = centennial
        self.hop2 = cascade
        self.addition_kwargs = [
            {
                'boil_time': 60.0,
                'weight': 0.57,
                'utilization_cls_kwargs': {'units': IMPERIAL_UNITS},
            },
            {
                'boil_time': 5.0,
                'weight': 0.76,
                'utilization_cls_kwargs': {'units': IMPERIAL_UNITS},
            },
        ]

        # Additions
        self.hop_addition1 = HopAddition(self.hop1,
                                         **self.addition_kwargs[0])
        self.hop_addition2 = HopAddition(self.hop2,
                                         **self.addition_kwargs[1])

        # Define Hops
        self.plato = 14.0
        self.sg = plato_to_sg(self.plato)
        self.target_ibu = 40.0
        self.final_volume = 5.0

    def test_change_units(self):
        self.assertEquals(self.hop_addition1.units, IMPERIAL_UNITS)
        ha = self.hop_addition1.change_units()
        self.assertEquals(ha.units, SI_UNITS)
        self.assertEquals(ha.utilization_cls.units, SI_UNITS)
        ha = ha.change_units()
        self.assertEquals(ha.units, IMPERIAL_UNITS)
        self.assertEquals(ha.utilization_cls.units, IMPERIAL_UNITS)

    def test_str(self):
        out = str(self.hop_addition1)
        self.assertEquals(
            out, 'Centennial, alpha 0.14%, 0.57 oz, 60.0 min, pellet')

    def test_repr(self):
        out = repr(self.hop_addition1)
        self.assertEquals(
                out, "HopAddition(Hop('centennial', percent_alpha_acids=0.14), weight=0.57, boil_time=60.0, hop_type='pellet', utilization_cls=HopsUtilizationGlennTinseth, utilization_cls_kwargs={'units': 'imperial'}, units='imperial')")  # nopep8

    def test_to_dict(self):
        out = self.hop_addition1.to_dict()
        expected = {'name': 'centennial',
                    'data': {'percent_alpha_acids': 0.14},
                    'weight': 0.57,
                    'boil_time': 60.0,
                    'hop_type': 'pellet',
                    'utilization_cls': 'Glenn Tinseth',
                    'utilization_cls_kwargs': {'units': 'imperial'},
                    'units': 'imperial',
                    }
        self.assertEquals(out, expected)

    def test_to_json(self):
        out = self.hop_addition1.to_json()
        expected = '{"boil_time": 60.0, "data": {"percent_alpha_acids": 0.14}, "hop_type": "pellet", "name": "centennial", "units": "imperial", "utilization_cls": "Glenn Tinseth", "utilization_cls_kwargs": {"units": "imperial"}, "weight": 0.57}'  # nopep8
        self.assertEquals(out, expected)

    def test_validate(self):
        data = self.hop_addition1.to_dict()
        HopAddition.validate(data)

    def test_format(self):
        out = self.hop_addition1.format()
        msg = textwrap.dedent("""\
                centennial Addition
                -----------------------------------
                Hop Type:     pellet
                AA %:         0.14 %
                Weight:       0.57 oz
                Boil Time:    60.0 min""")
        self.assertEquals(out, msg)

    def test_get_hops_weight(self):
        hops_weight = self.hop_addition1.get_hops_weight(
            self.sg,
            self.target_ibu,
            self.final_volume,
            0.95)
        self.assertEquals(round(hops_weight, 2), 0.84)

        hops_weight = self.hop_addition2.get_hops_weight(
            self.sg,
            self.target_ibu,
            self.final_volume,
            0.05)
        self.assertEquals(round(hops_weight, 2), 0.44)

    def test_get_hops_weight_metric(self):
        ha1 = self.hop_addition1.change_units()
        hops_weight = ha1.get_hops_weight(
            self.sg,
            self.target_ibu,
            self.final_volume * LITER_PER_GAL,
            0.95)
        self.assertEquals(round(hops_weight, 2), 23684.97)

        ha2 = self.hop_addition2.change_units()
        hops_weight = ha2.get_hops_weight(
            self.sg,
            self.target_ibu,
            self.final_volume * LITER_PER_GAL,
            0.05)
        self.assertEquals(round(hops_weight, 2), 12506.15)

    def test_get_alpha_acid_units(self):
        out = self.hop_addition1.get_alpha_acid_units()
        self.assertEquals(round(out, 2), 7.98)

        ha = self.hop_addition1.change_units()
        out = ha.get_alpha_acid_units()
        self.assertEquals(round(out, 2), 7.98)



import math

from ..constants import GAL_PER_LITER
from ..constants import IMPERIAL_TYPES
from ..constants import IMPERIAL_UNITS
from ..constants import LITER_PER_GAL
from ..constants import OZ_PER_G
from ..constants import SI_TYPES
from ..constants import SI_UNITS
from ..validators import validate_units
from .sugar import plato_to_sg
from .sugar import sg_to_gu
from .sugar import sg_to_plato


__all__ = [
    'PITCH_RATE_MAP',
    'pitch_rate_conversion',
    'YeastModel',
    'KaiserYeastModel',
    'WhiteYeastModel',
]


"""
1 billion cells growth per gram of extract (B/g) =
    13.3 Million cells / (ml * P)

Ale:    0.75 M / ml / P = 0.71  B / G / SG
Lager:  1.50 M / ml / P = 1.42  B / G / SG
Hybrid: 1.00 M / ml / P = 0.948 B / G / SG
"""

# Pitch rate in M / ml / P
PITCH_RATE_MAP = {
    'MFG Recommended  (Ale, fresh yeast only)': 0.35,
    'MFG Recommended+ (Ale, fresh yeast only)': 0.55,
    'Pro Brewer (Ale, LG)': 0.75,
    'Pro Brewer (Ale)': 1.0,
    'Pro Brewer (Ale, HG)': 1.25,
    'Pro Brewer (Lager, LG)': 1.5,
    'Pro Brewer (Lager)': 1.75,
    'Pro Brewer (Lager, HG)': 2.0,
}


def pitch_rate_conversion(pitch_rate, units=IMPERIAL_UNITS):
    """
    Pitch Rate Conversion

    Input should be given in:
    Imperial: B / (Gal * GU)
    SI:       B / (L * P)

    Note: 1 M / (ml * P) == 1B / (L * P)
    """
    plato_per_gu = sg_to_gu(plato_to_sg(1))
    if units == IMPERIAL_UNITS:
        return pitch_rate * GAL_PER_LITER * plato_per_gu
    elif units == SI_UNITS:
        return pitch_rate * LITER_PER_GAL / plato_per_gu


class YeastModel(object):
    METHOD_TO_GROWTH_ADJ = {
        'no agitation': 0.0,
        'shaking': 0.0,
        'stir plate': 0.0,
    }

    def __init__(self, method, units=IMPERIAL_UNITS):
        if method not in self.METHOD_TO_GROWTH_ADJ.keys():
            raise Exception("Method '{}' not allowed for yeast model".format(
                method))
        self.method = method
        self.adjustment = self.METHOD_TO_GROWTH_ADJ[method]
        self.set_units(units)

    def set_units(self, units):
        self.units = validate_units(units)
        if self.units == IMPERIAL_UNITS:
            self.types = IMPERIAL_TYPES
        elif self.units == SI_UNITS:
            self.types = SI_TYPES

    def get_inoculation_rate(self, growth_rate):
        raise NotImplementedError

    def get_growth_rate(self, inoculation_rate):
        raise NotImplementedError

    def get_viability(self, days_since_manufacture):
        """
        Yeast viability drops 21% each month or 0.7% per day from the date of
        manufacture.  Assume linear change.
        """
        viability = 1.0 - days_since_manufacture * (0.21 / 30.0)
        if viability < 0:
            return 0.0
        return viability

    def get_yeast_pitch_rate(self,
                             original_gravity=1.050,
                             final_volume=5.0,
                             target_pitch_rate=1.42,
                             yeast_type='liquid',
                             cells_per_pack=100,
                             num_packs=1,
                             days_since_manufacture=30):
        """
        Determine yeast pitch rate

        original_gravity  - specific gravity of original beer
        final_volume      - volume of the batch post fermentation
        target_pitch_rate - million cells / (ml * degP)
        yeast_type        - liquid, dry
        cells_per_pack    - Billions of cells
        num_packs         - how many in units
        days_since_manufacture - the older the yeast the less viable
        units             - imperial, metric

        Yeast Viability: lose 20% viability / month or 0.66% / day

        Imperial: B / Gal / GU
        Metric:   M / ml  / Plato

        Sources:
        - http://beersmith.com/blog/2011/01/10/yeast-starters-for-home-brewing-beer-part-2/
        """  # nopep8
        viability = self.get_viability(days_since_manufacture)

        cells = cells_per_pack * num_packs * viability

        if self.units == IMPERIAL_UNITS:
            modifier = sg_to_gu(original_gravity)
        elif self.units == SI_UNITS:
            modifier = sg_to_plato(original_gravity)
        pitch_rate_as_is = cells / final_volume / modifier
        pitch_rate_cells = target_pitch_rate * final_volume * modifier

        if cells <= 0.0:
            required_growth_rate = 0.0
        else:
            required_growth_rate = pitch_rate_cells / cells

        return {'original_gravity': original_gravity,
                'final_volume': final_volume,
                'target_pitch_rate': target_pitch_rate,
                'viability': round(viability, 2),
                'cells': round(cells, 2),
                'pitch_rate_as_is': round(pitch_rate_as_is, 2),
                'pitch_rate_cells': round(pitch_rate_cells, 2),
                'cells_needed': round(pitch_rate_cells - cells, 2),
                'required_growth_rate': round(required_growth_rate, 2),
                'units': self.units,
                }

    def get_starter_volume(self,
                           available_cells,
                           starter_volume=2.0 * GAL_PER_LITER,
                           original_gravity=1.036):
        """
        Calculate the number of cells given a stater volume and gravity
        """
        GPL = 2.845833  # g/P/L grams of extract per point of gravity per liter of starter  # nopep8
        dme = GPL * sg_to_gu(original_gravity) * starter_volume  # in grams
        if self.units == IMPERIAL_UNITS:
            inoculation_rate = available_cells / (starter_volume * LITER_PER_GAL)  # nopep8
            dme = dme * OZ_PER_G * LITER_PER_GAL
        elif self.units == SI_UNITS:
            inoculation_rate = available_cells / starter_volume
        growth_rate = self.get_growth_rate(inoculation_rate)
        end_cell_count = available_cells * (growth_rate + 1)

        return {'available_cells': round(available_cells, 2),
                'starter_volume': round(starter_volume, 2),
                'original_gravity': original_gravity,
                'dme': round(dme, 2),
                'inoculation_rate': round(inoculation_rate, 2),
                'growth_rate': round(growth_rate, 2),
                'end_cell_count': round(end_cell_count, 2),
                'units': self.units,
                }

    def get_resulting_pitch_rate(self,
                                 starter_cell_count,
                                 original_gravity=1.036,
                                 final_volume=5.0):
        if self.units == IMPERIAL_UNITS:
            modifier = sg_to_gu(original_gravity)
        elif self.units == SI_UNITS:
            modifier = sg_to_plato(original_gravity)
        pitch_rate = starter_cell_count / final_volume / modifier
        return pitch_rate


class KaiserYeastModel(YeastModel):
    """
    Kaiser Yeast Model

    Only works for Stir Plage Growth

    Sources:
    - http://braukaiser.com/blog/blog/2012/11/03/estimating-yeast-growth/
    """
    METHOD_TO_GROWTH_ADJ = {
        'stir plate': 0.0,
    }

    def __init__(self, method='stir plate', units=IMPERIAL_UNITS):
        return super(KaiserYeastModel, self).__init__(method, units=units)

    def get_inoculation_rate(self, growth_rate):
        if 0 < growth_rate < 1.4:
            return (2.33 - growth_rate) / 0.67
        elif 1.4 <= growth_rate:
            return 1.4

    def get_growth_rate(self, initial_cells):
        """
        initial_cells - Billion / gram extract (B/g)
        """
        if initial_cells < 1.4:
            return 1.4 + self.adjustment
        elif 1.4 <= initial_cells < 3.5:
            return 2.33 - 0.67 * initial_cells + self.adjustment
        else:
            return 0.0 + self.adjustment


class WhiteYeastModel(YeastModel):
    """
    Sources:
    - http://www.brewersfriend.com/yeast-pitch-rate-and-starter-calculator/
    - White, Chris, and Jamil Zainasheff. Yeast: The Practical Guide to Beer
      Fermentation. Boulder, CO: Brewers Publications, 2010. 139-44. Print.
    """

    # Linear Regression Least Squares
    INOCULATION_CONST = [-0.999499, 12.547938, -0.459486]
    METHOD_TO_GROWTH_ADJ = {
        'no agitation': 0.0,
        'shaking': 0.5,
        'stir plate': 1.0,
    }

    def __init__(self, method='no agitation', units=IMPERIAL_UNITS):
        return super(WhiteYeastModel, self).__init__(method, units=units)

    def get_inoculation_rate(self, growth_rate):
        a, b, c = self.INOCULATION_CONST
        return 10 ** (math.log10(b / (growth_rate - a)) / -c)

    def get_growth_rate(self, inoculation_rate):
        """
        initial_cells - Billion / gram extract (B/g)

        G = (12.54793776 * x^-0.4594858324) - 0.9994994906
        """  # nopep8
        # if inoculation_rate > 200:
        #     raise Exception("Yeast will not grow at more than 200 M/ml")
        a, b, c = self.INOCULATION_CONST
        growth_rate = a + b * inoculation_rate ** c + self.adjustment
        # if growth_rate > 6:
        #     raise Exception("Model does not allow for growth greater than 6")
        return growth_rate





__all__ = [
    'fahrenheit_to_celsius',
    'celsius_to_fahrenheit',
]


def fahrenheit_to_celsius(temp):
    """
    Convert degrees Fahrenheit  to degrees Celsius
    """
    return (temp - 32.0) / 1.8


def celsius_to_fahrenheit(temp):
    """
    Convert degrees Celsius to degrees Fahrenheit
    """
    return(temp * 1.8) + 32.0



import unittest

from brew.utilities.temperature import celsius_to_fahrenheit
from brew.utilities.temperature import fahrenheit_to_celsius


class TestTemperatureUtilities(unittest.TestCase):

    def test_celsius_to_fahrenheit(self):
        ftemp = celsius_to_fahrenheit(100.0)
        self.assertEquals(ftemp, 212.0)
        ftemp = celsius_to_fahrenheit(0.0)
        self.assertEquals(ftemp, 32.0)
        ftemp = celsius_to_fahrenheit(-40.0)
        self.assertEquals(ftemp, -40.0)

    def test_fahrenheit_to_celsius(self):
        ctemp = fahrenheit_to_celsius(212.0)
        self.assertEquals(ctemp, 100.0)
        ctemp = fahrenheit_to_celsius(32.0)
        self.assertEquals(ctemp, 0.0)
        ctemp = fahrenheit_to_celsius(-40.0)
        self.assertEquals(ctemp, -40.0)



#! /usr/bin/env python

"""
Oak Barrel Winecraft

Kolsch Ale

Used by permission of Oak Barrel Winecraft. All rights reserved.

You can purchase this kit at their store:

- http://oakbarrel.com/

Original Stats:

OG:      1.045
FG:      1.008 - 1.012
ADF:     72 % - 78 %
IBU:
Color:
Alcohol: 4.3% ABV
Boil:    60 min
Pre-Boil Volume:  2.5 G
Pre-Boil Gravity: 
"""  # nopep8

import os

from brew.parsers import JSONDataLoader
from brew.parsers import parse_recipe


def main():

    recipe = {
        'name': "Kolsch Ale (Extract)",
        'start_volume': 2.5,
        'final_volume': 5.0,
        'grains': [
            {'name': 'Pilsner Liquid Extract',
             'weight': 3.25,
             'grain_type': 'lme'},
            {'name': 'Munich Liquid Extract',
             'data': {
                 'color': 10.0,
                 'ppg': 36,
             },
             'weight': 3.25,
             'grain_type': 'lme'},
            {'name': 'White Wheat Malt',
             'weight': 0.25,
             'grain_type': 'specialty'},
            {'name': 'Caramel Crystal Malt 10l',
             'weight': 0.25,
             'grain_type': 'specialty'},
        ],
        'hops': [
            {'name': 'Vanguard',
             'weight': 1.0,
             'boil_time': 60.0},
            {'name': 'hersbrucker',
             'weight': 1.0,
             'boil_time': 0.0},
        ],
        'yeast': {
            'name': 'White Labs Wlp029',
        },
        'data': {
            'percent_brew_house_yield': 0.70,
            'units': 'imperial',
        },
    }

    data_dir = os.path.abspath(os.path.join(os.getcwd(), 'data/'))
    loader = JSONDataLoader(data_dir)
    beer = parse_recipe(recipe, loader)
    print(beer.format())


if __name__ == "__main__":
    main()




import argparse
import sys
import textwrap

from brew.utilities.sugar import brix_to_plato
from brew.utilities.sugar import brix_to_sg
from brew.utilities.sugar import plato_to_brix
from brew.utilities.sugar import plato_to_sg
from brew.utilities.sugar import sg_to_brix
from brew.utilities.sugar import sg_to_plato


def get_sugar_conversion(brix_in, plato_in, sg_in, sugar_out):
    """
    Convert one sugar unit to another or print all.

    brix_in - Degrees Brix Input
    plato_in - Degrees Plato Input
    sg_in - Specific Gravity Input
    sugar_out - Type of conversion ('b', 'p', 's', or None)
    """
    brix, plato, sg = 0.0, 0.0, 0.0
    if brix_in:
        brix = brix_in
        plato = brix_to_plato(brix_in)
        sg = brix_to_sg(brix_in)
    elif plato_in:
        brix = plato_to_brix(plato_in)
        plato = plato_in
        sg = plato_to_sg(plato_in)
    elif sg_in:
        brix = sg_to_brix(sg_in)
        plato = sg_to_plato(sg_in)
        sg = sg_in

    brix = round(brix, 1)
    plato = round(plato, 1)
    sg = round(sg, 3)
    if sugar_out and sugar_out in ['b', 'p', 's']:
        if sugar_out == 'b':
            return brix
        elif sugar_out == 'p':
            return plato
        elif sugar_out == 's':
            return sg
    else:
        out = textwrap.dedent("""\
        SG\tPlato\tBrix
        {:0.3f}\t{:0.1f}\t{:0.1f}""".format(sg, plato, brix))
        return out


def get_parser():
    parser = argparse.ArgumentParser(description='Sugar Conversion')
    parser.add_argument('-b', '--brix', metavar='B', type=float,
                        help='Degrees Brix')
    parser.add_argument('-p', '--plato', metavar='P', type=float,
                        help='Degrees Plato')
    parser.add_argument('-s', '--sg', metavar='S', type=float,
                        help='Specific Gravity')
    parser.add_argument('-o', '--out', metavar='O', type=str,
                        help='Desired Output (b, p, s accepted)')
    return parser


def main():
    parser = get_parser()
    args = parser.parse_args()

    if sum(bool(arg) for arg in [args.brix, args.plato, args.sg]) != 1:
        print("Must provide only one of Brix, Plato or Specific Gravity")
        sys.exit(1)
    print(get_sugar_conversion(args.brix, args.plato, args.sg, args.out))



import unittest

from brew.constants import GRAIN_TYPE_CEREAL
from brew.constants import HOP_TYPE_PELLET
from brew.validators import validate_grain_type
from brew.validators import validate_hop_type
from brew.validators import validate_percentage
from brew.validators import validate_units
from brew.validators import validate_required_fields
from brew.validators import validate_optional_fields


class TestValidators(unittest.TestCase):

    def test_validate_grain_type(self):
        out = validate_grain_type(GRAIN_TYPE_CEREAL)
        self.assertEqual(out, GRAIN_TYPE_CEREAL)

    def test_validate_grain_type_raises(self):
        with self.assertRaises(Exception):
            validate_grain_type('bad grain type')

    def test_validate_hop_type(self):
        out = validate_hop_type(HOP_TYPE_PELLET)
        self.assertEqual(out, HOP_TYPE_PELLET)

    def test_validate_hop_type_raises(self):
        with self.assertRaises(Exception):
            validate_hop_type('bad hop type')

    def test_validate_percentage_pass(self):
        out = validate_percentage(0.97)
        self.assertEquals(out, 0.97)

    def test_validate_percentage_raises(self):
        with self.assertRaises(Exception):
            validate_percentage(1.01)

        with self.assertRaises(Exception):
            validate_percentage(-0.01)

    def test_validate_units(self):
        out = validate_units('metric')
        self.assertEquals(out, 'metric')

    def test_validate_units_raises(self):
        with self.assertRaises(Exception):
            validate_units('bad')

    def test_validate_required_fields(self):
        data = {'required': 'data'}
        required_fields = [('required', str)]
        validate_required_fields(data, required_fields)

    def test_validate_required_fields_missing_field_raises(self):
        data = {'missing': 'data'}
        required_fields = [('required', str)]
        with self.assertRaises(Exception):
            validate_required_fields(data, required_fields)

    def test_validate_required_fields_wrong_field_type_raises(self):
        data = {'required': 'data'}
        required_fields = [('required', int)]
        with self.assertRaises(Exception):
            validate_required_fields(data, required_fields)

    def test_validate_optional_fields(self):
        data = {'data': {'optional': 'data'}}
        optional_fields = [('optional', str)]
        validate_optional_fields(data, optional_fields)

    def test_validate_optional_fields_extra_data(self):
        data = {'data': {'extra': 'data'}}
        optional_fields = [('optional', str)]
        validate_optional_fields(data, optional_fields)

    def test_validate_optional_fields_missing_data_field(self):
        data = {'missing': {'optional': 'data'}}
        optional_fields = [('optional', str)]
        validate_optional_fields(data, optional_fields)

    def test_validate_optional_fields_wrong_field_type_raises(self):
        data = {'data': {'optional': 'data'}}
        optional_fields = [('optional', int)]
        with self.assertRaises(Exception):
            validate_optional_fields(data, optional_fields)



import json
import textwrap

from .validators import validate_optional_fields
from .validators import validate_percentage
from .validators import validate_required_fields


__all__ = ['Yeast']


class Yeast(object):
    """
    A representation of a type of Yeast as added to a Recipe.
    """

    def __init__(self, name,
                 percent_attenuation=0.75):
        """
        Percent Attenuation - The percentage the yeast is expected to
            attenuate the sugar in the beer to create alcohol.
        """
        self.name = name
        self.percent_attenuation = validate_percentage(percent_attenuation)

    def __str__(self):
        return "{0}, attenuation {1}%".format(self.name.capitalize(),
                                              self.percent_attenuation)

    def __repr__(self):
        out = "{0}('{1}'".format(type(self).__name__, self.name)
        if self.percent_attenuation:
            out = "{0}, percent_attenuation={1}".format(
                out, self.percent_attenuation)
        out = "{0})".format(out)
        return out

    def to_dict(self):
        return {'name': self.name,
                'data': {
                    'percent_attenuation': self.percent_attenuation,
                },
                }

    def to_json(self):
        return json.dumps(self.to_dict(), sort_keys=True)

    @classmethod
    def validate(cls, yeast_data):
        required_fields = [('name', str),
                           ]
        optional_fields = [('percent_attenuation', float),
                           ]
        validate_required_fields(yeast_data, required_fields)
        validate_optional_fields(yeast_data, optional_fields)

    def format(self):
        msg = textwrap.dedent("""\
                {name} Yeast
                -----------------------------------
                Attenuation:  {data[percent_attenuation]} %""".format(
            **self.to_dict()))
        return msg



#! /usr/bin/env python

import os

from brew.parsers import JSONDataLoader
from brew.parsers import parse_recipe


def main():

    recipe = {
        'name': 'Pale Ale',
        'start_volume': 7.0,
        'final_volume': 5.0,
        'grains': [
            {'name': 'pale malt 2-row us',
             'data': {
                 'color': 1.8,
                 'ppg': 37,
             },
             'weight': 13.96},
            {'name': 'caramel crystal malt 20l',
             'data': {
                 'color': 20.0,
                 'ppg': 35,
             },
             'weight': 0.78},
        ],
        'hops': [
            {'name': 'centennial',
             'data': {
                 'percent_alpha_acids': 0.14,
             },
             'weight': 0.57,
             'boil_time': 60.0},
            {'name': 'cascade us',
             'data': {
                 'percent_alpha_acids': 0.07,
             },
             'weight': 0.76,
             'boil_time': 5.0},
        ],
        'yeast': {
            'name': 'Wyeast 1056',
            'data': {
                'percent_attenuation': 0.75,
            },
        },
        'data': {
            'percent_brew_house_yield': 0.70,
            'units': 'imperial',
        },
    }

    data_dir = os.path.abspath(os.path.join(os.getcwd(), 'data/'))
    loader = JSONDataLoader(data_dir)
    beer = parse_recipe(recipe, loader)
    print(beer.format())


if __name__ == "__main__":
    main()
